from pathlib import Path

from django.db import models
from django.utils.timezone import now
from django_enumfield.enum import EnumField
from markdown import markdown

from mailinglist.conf import hookset, settings
from mailinglist.enum import SubmissionStatusEnum, SubscriptionStatusEnum


class MailingList(models.Model):
    """The fundamental model for sending messages. A mailing list can be
    subscribed to by users and is required for composing (and therefore
    sending) messages"""

    name = models.CharField(max_length=128)
    slug = models.SlugField(db_index=True, unique=True)
    email = models.EmailField(help_text="Sender e-mail")
    sender = models.CharField(max_length=200, help_text="Sender name")
    visible = models.BooleanField(default=True, db_index=True)
    send_html = models.BooleanField(
        default=True,
        help_text="Whether or not to send HTML versions of e-mails.",
    )

    def __str__(self):
        return self.name

    @property
    def sender_tag(self):
        """Mailing list sender formatted in the standard way for email addresses::
        "FirstName LastName" <email>
        """
        return f'"{self.sender}" <{self.email}>'

    @property
    def published_messages(self):
        """All messages that have been published to this mailing list prior to now"""
        return self.messages.filter(submission__published__lte=now()).distinct()


class GlobalDeny(models.Model):
    """Users with instances here will not receive mailing list messages."""

    user = models.OneToOneField(
        settings.MAILINGLIST_USER_MODEL,
        on_delete=models.CASCADE,
        related_name="mailinglist_deny",
    )
    created = models.DateTimeField(auto_now_add=True, editable=False)


class Subscription(models.Model):
    """The means by which a user subscribes to receive mailing list messages.
    **Do not create instances of this yourself, use the methods in
    ``mailinglist.services.SubscriptionService`` for managing subscriptions!**"""

    user = models.ForeignKey(
        settings.MAILINGLIST_USER_MODEL,
        on_delete=models.CASCADE,
        related_name="subscriptions",
    )
    # token used for verification of subscription events
    token = models.CharField(max_length=45)
    mailing_list = models.ForeignKey(
        MailingList,
        on_delete=models.CASCADE,
        related_name="subscriptions",
        null=True,
    )
    status = EnumField(SubscriptionStatusEnum)

    def __str__(self):
        return f"{self.user} on {self.mailing_list}"

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["user", "mailing_list"], name="unique_user_per_mailing_list"
            )
        ]


class SubscriptionChange(models.Model):
    """Tracks changes to subscriptions so that subscriptions/unsubscriptions
    can be audited. Instances of this model are managed by
    ``mailinglist.services.SubscriptionService``."""

    subscription = models.ForeignKey(Subscription, on_delete=models.CASCADE)
    from_status = EnumField(SubscriptionStatusEnum, null=True, blank=True)
    to_status = EnumField(SubscriptionStatusEnum)
    changed = models.DateTimeField(auto_now_add=True)


class Message(models.Model):
    """The messages which will be sent to subscribed users. Comprised of
    ``MessagePart`` and ``MessageAttachment`` instances."""

    title = models.CharField(max_length=128)
    slug = models.SlugField()
    mailing_list = models.ForeignKey(
        MailingList, on_delete=models.CASCADE, related_name="messages"
    )
    created = models.DateTimeField(auto_now_add=True, editable=False)
    modified = models.DateTimeField(auto_now=True, editable=False)

    def __str__(self):
        return self.title

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["slug", "mailing_list"], name="unique_slug_per_mailing_list"
            )
        ]


class MessagePart(models.Model):
    """The text content of a ``Message``."""

    message = models.ForeignKey(
        Message, on_delete=models.CASCADE, related_name="message_parts"
    )
    heading = models.CharField(max_length=128)
    order = models.PositiveSmallIntegerField()
    text = models.TextField()
    # TODO: images!

    @property
    def html_text(self):
        return markdown(self.text)

    class Meta:
        ordering = ["order"]
        constraints = [
            models.UniqueConstraint(
                fields=["message", "order"], name="unique_order_per_message"
            )
        ]


def attachment_upload_to(instance, filename):
    return Path(
        "mailinglist-static",
        "attachments",
        str(instance.message.pk),
        filename,
    )


def hookset_validation_wrapper(value):
    # Model field validators are referenced in migration files. This wrapper
    #  allows the hookset to be dynamic without causing chaos in the migration
    #  file.
    hookset.message_attachment_file_validator(value)


class MessageAttachment(models.Model):
    """The file content of a ``Message``"""

    message = models.ForeignKey(
        Message, on_delete=models.CASCADE, related_name="attachments"
    )
    file = models.FileField(
        upload_to=attachment_upload_to,
        verbose_name="attachment",
        validators=[hookset_validation_wrapper],
    )
    filename = models.CharField(max_length=256, null=False)

    def __str__(self):
        return f"{self.filename} on {self.message.title}"

    def save(self, **kwargs):
        # The file name may be mangled by name collisions upon save,
        #   capture the original filename prior to the first save.
        if not self.pk:
            self.filename = str(self.file)
        super().save(**kwargs)


class Submission(models.Model):
    """The means by which ``Message`` instances are published and sent."""

    message = models.OneToOneField(
        Message, on_delete=models.PROTECT, related_name="submission"
    )
    exclude = models.ManyToManyField(
        Subscription,
        blank=True,
        help_text=(
            "Subscriptions to exclude from mailing, all other "
            "subscribers will receive message"
        ),
    )
    published = models.DateTimeField(null=True, blank=True)
    status = EnumField(SubmissionStatusEnum)
    sendings = models.ManyToManyField(
        Subscription, through="Sending", related_name="sendings"
    )

    def __str__(self):
        return f"{self.message} to {self.message.mailing_list}"


class Sending(models.Model):
    """Tracks the sending of each ``Submission`` to each individual
    ``Subscription``. Provided for audit, as well as to allow interruped
    send jobs to be resumed without double-sending to any users."""

    submission = models.ForeignKey(Submission, on_delete=models.PROTECT)
    subscription = models.ForeignKey(Subscription, on_delete=models.PROTECT)
    sent = models.DateTimeField(auto_now_add=True)



import logging

logger = logging.getLogger(__name__)

import io
from csv import DictReader

from django import forms
from django.apps import apps
from django.conf import settings
from django.core.exceptions import ValidationError
from django.core.validators import validate_email

from mailinglist.models import Subscription


class AddressList:
    """List with unique addresses."""

    def __init__(self, mailing_list, ignore_errors=False):
        self.mailing_list = mailing_list
        self.ignore_errors = ignore_errors
        self.addresses = {}

    class OverLongEmailException(Exception):
        pass

    class InvalidEmailException(Exception):
        pass

    def _validate_email(self, email):
        try:
            email = check_field(
                field_name="email", value=email.lower(), ignore_errors=False
            )
        except forms.ValidationError as e:
            logger.warning(str(e))
            raise self.OverLongEmailException

        try:
            validate_email(email)
        except ValidationError:
            logger.warning(f"Entry '{email}' does not contain a valid e-mail address")
            raise self.InvalidEmailException

    def add(
        self,
        *,
        email,
        first_name=None,
        last_name=None,
        location="unknown location",
        **kwargs,
    ):
        """Add name to list."""

        logger.debug(f"Going to add {first_name} {last_name} <{email}>")
        try:
            self._validate_email(email)
        except self.OverLongEmailException:
            if not self.ignore_errors:
                raise forms.ValidationError(
                    f"Entry '{first_name} {last_name}' at {location} has an "
                    "overlong email address."
                )
            return
        except self.InvalidEmailException:
            if not self.ignore_errors:
                raise forms.ValidationError(
                    f"Entry '{first_name} {last_name}' at {location} does "
                    "not contain a valid email address."
                )
            return

        first_name = check_field(
            field_name="first_name", value=first_name, ignore_errors=self.ignore_errors
        )
        last_name = check_field(
            field_name="last_name", value=last_name, ignore_errors=self.ignore_errors
        )

        if email in self.addresses:
            logger.warning(
                f"Entry '{first_name} {last_name}' contains a duplicate entry "
                f"at {location}."
            )

            if not self.ignore_errors:
                raise forms.ValidationError(
                    f"The address file contains duplicate entries for '{email}'."
                )
            return

        if subscription_exists(self.mailing_list, email):
            logger.warning(f"Entry '{email}' is already subscribed to at {location}.")

            if not self.ignore_errors:
                raise forms.ValidationError("Some entries are already subscribed to.")
            return

        self.addresses[email] = {
            "first_name": first_name,
            "last_name": last_name,
            "email": email,
        }


def subscription_exists(mailing_list, email):
    """
    Return wheter or not a subscription exists.
    """
    qs = Subscription.objects.filter(
        mailing_list__id=mailing_list.id, user__email__iexact=email
    )

    return qs.exists()


def check_field(*, field_name, value, ignore_errors=False):
    """
    Check (length of) value address.
    """

    logger.debug(f"Checking {field_name} '{value}'")
    if value is None:
        return None

    max_length = (
        apps.get_model(settings.MAILINGLIST_USER_MODEL)
        ._meta.get_field(field_name)
        .max_length
    )

    # Get rid of leading/trailing spaces
    value = value.strip()

    if len(value) <= max_length or ignore_errors:
        return value[:max_length]
    else:
        raise forms.ValidationError(
            f"{field_name} value '{value}' too long, maximum length is "
            f"{max_length} characters."
        )


def parse_csv(csv_file, mailing_list, ignore_errors=False):
    """
    Parse addresses from CSV file-object into mailing list.

    Returns a dictionary mapping email addresses into Subscription objects.
    """
    address_list = AddressList(mailing_list, ignore_errors)
    _csv_file = io.StringIO(csv_file.read().decode())
    reader = DictReader(_csv_file)
    for idx, row in enumerate(reader):
        address_list.add(**row, location=f"line {idx}")

    return address_list.addresses



from unittest.mock import patch, Mock
import pytest
from mailinglist.hooks import MailinglistDefaultHookset
from django.core.exceptions import ValidationError
from django.test import override_settings


class FileNameyThingy:
    def __init__(self, name):
        self.name = name


class TestMailinglistDefaultHookset:
    def test_create_user(self, db):
        user = MailinglistDefaultHookset().create_user(
            email="test@email.com",
            first_name="testy",
            last_name="mctestface",
        )
        assert user.email == "test@email.com"
        assert user.first_name == "testy"
        assert user.last_name == "mctestface"

    def test_create_user_exists(self, user):
        _user = MailinglistDefaultHookset().create_user(
            email=user.email,
            first_name=user.first_name + "asdfsadf",
            last_name=user.last_name + "zxcvzxcv",
        )
        assert _user.pk == user.pk
        assert _user.first_name == user.first_name  # didn't get updated
        assert _user.last_name == user.last_name  # didn't get updated

    @override_settings(MAILINGLIST_USER_MODEL="custom.override")
    @patch("django.apps.apps.get_model")
    def test_create_user_other_model(self, p_get_model):
        _user_model = Mock()
        _user_model.objects.get.return_value = "asdf"
        p_get_model.return_value = _user_model
        user = MailinglistDefaultHookset().create_user(
            email="test@email.com",
            first_name="testy",
            last_name="mctestface",
        )
        p_get_model.assert_called_once_with("custom.override")
        assert user == "asdf"

    @patch("mailinglist.hooks.EmailMultiAlternatives")
    def test_send_message(self, p_email_alternatives, message_attachment):
        _message = Mock()
        p_email_alternatives.return_value = _message
        MailinglistDefaultHookset().send_message(
            to="someone@email.com",
            body="good strong body here!",
            subject="[mailinglist] subject or whatever",
            from_email="loving@it.com",
            html_body="good <strong>strong</strong> body here!",
            attachments=[message_attachment],
            headers={"header": "yes"},
        )
        p_email_alternatives.assert_called_once_with(
            to="someone@email.com",
            body="good strong body here!",
            subject="[mailinglist] subject or whatever",
            from_email="loving@it.com",
            headers={"header": "yes"},
        )
        _message.attach_file.assert_called_once_with(message_attachment.file.path)
        _message.attach_alternative.assert_called_once_with(
            "good <strong>strong</strong> body here!", "text/html"
        )
        _message.send.assert_called_once_with()

    @patch("mailinglist.hooks.EmailMultiAlternatives")
    def test_send_message_no_attachments(self, p_email_alternatives):
        _message = Mock()
        p_email_alternatives.return_value = _message
        MailinglistDefaultHookset().send_message(
            to="someone@email.com",
            body="good strong body here!",
            subject="[mailinglist] subject or whatever",
            from_email="loving@it.com",
            html_body="good <strong>strong</strong> body here!",
            headers={"header": "yes"},
        )
        p_email_alternatives.assert_called_once_with(
            to="someone@email.com",
            body="good strong body here!",
            subject="[mailinglist] subject or whatever",
            from_email="loving@it.com",
            headers={"header": "yes"},
        )
        _message.attach_file.assert_not_called()
        _message.attach_alternative.assert_called_once_with(
            "good <strong>strong</strong> body here!", "text/html"
        )
        _message.send.assert_called_once_with()

    @patch("mailinglist.hooks.EmailMultiAlternatives")
    def test_send_message_no_html(self, p_email_alternatives, message_attachment):
        _message = Mock()
        p_email_alternatives.return_value = _message
        MailinglistDefaultHookset().send_message(
            to="someone@email.com",
            body="good strong body here!",
            subject="[mailinglist] subject or whatever",
            from_email="loving@it.com",
            attachments=[message_attachment],
            headers={"header": "yes"},
        )
        p_email_alternatives.assert_called_once_with(
            to="someone@email.com",
            body="good strong body here!",
            subject="[mailinglist] subject or whatever",
            from_email="loving@it.com",
            headers={"header": "yes"},
        )
        _message.attach_file.assert_called_once_with(message_attachment.file.path)
        _message.attach_alternative.assert_not_called()
        _message.send.assert_called_once_with()

    def test_message_attachment_file_validator_bad(self):
        with pytest.raises(ValidationError):
            MailinglistDefaultHookset().message_attachment_file_validator(
                FileNameyThingy(name="file.exe")
            )

    def test_message_attachment_file_validator_jpg(self):
        MailinglistDefaultHookset().message_attachment_file_validator(
            FileNameyThingy("long/path/t/o/file.jpg")
        )

    def test_message_attachment_file_validator_jpeg(self):
        MailinglistDefaultHookset().message_attachment_file_validator(
            FileNameyThingy("long/path/t/o/file.jpeg")
        )

    def test_message_attachment_file_validator_pdf(self):
        MailinglistDefaultHookset().message_attachment_file_validator(
            FileNameyThingy("long/path/t/o/file.pdf")
        )



from unittest.mock import patch, Mock
from django.contrib.auth.models import AnonymousUser
from django.urls import reverse
from django.test import override_settings

from mailinglist import enum, models, views


def setup_view(view, request, *args, **kwargs):
    """Mimic as_view() returned callable, but returns view instance.

    args and kwargs are the same you would pass to ``reverse()``

    """
    view.request = request
    view.args = args
    view.kwargs = kwargs
    return view


class TestDetailFormView:
    # can't test the view class itself because it is too abstract
    @patch.object(views.SubscribeView, "get_object")
    def test_get(self, p_get_object, rf):
        request = rf.get("/fake-path")
        request.user = AnonymousUser()
        p_get_object.return_value = "great object"
        view = setup_view(views.SubscribeView(), request)
        view.get(request)
        assert view.object == "great object"


class TestSubscribeFormMixin:
    # can't test the mixin class itself because it is too abstract
    def test_initial_anon(self, rf):
        request = rf.get("/fake-path")
        request.user = AnonymousUser()
        view = setup_view(views.SubscribeView(), request)
        initial = view.get_initial()
        assert initial == {}

    def test_initial(self, rf, user):
        request = rf.get("/fake-path")
        request.user = user
        view = setup_view(views.SubscribeView(), request)
        initial = view.get_initial()
        assert initial == {
            "email": user.email,
            "first_name": user.first_name,
            "last_name": user.last_name,
        }


class TestGlobalDenyView:
    def test_get_success_url(self, rf):
        request = rf.get("/fake-path")
        request.user = AnonymousUser()
        view = setup_view(views.GlobalDenyView(), request)
        assert view.get_success_url() == "/mailinglist/global_deny/success/"

    def test_get_object(self, rf):
        request = rf.get("/fake-path")
        request.user = AnonymousUser()
        view = setup_view(views.GlobalDenyView(), request)
        assert view.get_object() is None


class TestSubscribeSuccessView:
    @override_settings(MAILINGLIST_CONFIRM_EMAIL_SUBSCRIBE=True)
    def test_get_context_data(self, rf):
        request = rf.get("/fake-path")
        request.user = AnonymousUser()
        view = setup_view(views.SubscribeSuccessView(), request)
        view.object = None
        assert view.get_context_data()["requires_confirmation"]

    @override_settings(MAILINGLIST_CONFIRM_EMAIL_SUBSCRIBE=False)
    def test_get_context_data_nay(self, rf):
        request = rf.get("/fake-path")
        request.user = AnonymousUser()
        view = setup_view(views.SubscribeSuccessView(), request)
        view.object = None
        assert not view.get_context_data()["requires_confirmation"]


class TestSubscriptionView:
    def test_get_success_url(self, rf, subscription):
        request = rf.get("/fake-path")
        request.user = AnonymousUser()
        view = setup_view(views.SubscriptionView(), request, token=subscription.token)
        assert (
            view.get_success_url()
            == f"/mailinglist/subscriptions/{subscription.token}/"
        )

    def test_form_valid(self, rf, subscription):
        request = rf.get("/fake-path")
        request.user = AnonymousUser()
        form = Mock()
        view = setup_view(views.SubscriptionView(), request, token=subscription.token)
        view.form_valid(form)
        form.save.assert_called_once()

    def test_get_form_kwargs(self, rf, subscription):
        request = rf.get("/fake-path")
        request.user = AnonymousUser()
        view = setup_view(views.SubscriptionView(), request, token=subscription.token)
        assert view.get_form_kwargs()["user"] == subscription.user


class TestSubscribeView:
    def test_post_form_empty(self, client, mailing_list):
        assert not models.Subscription.objects.all().exists()
        response = client.post(
            reverse(
                "mailinglist:subscribe", kwargs={"mailing_list_slug": mailing_list.slug}
            ),
        )
        assert not models.Subscription.objects.all().exists()
        assert not response.context["form"].is_valid()

    def test_post(self, client, mailing_list):
        assert not models.Subscription.objects.all().exists()
        client.post(
            reverse(
                "mailinglist:subscribe", kwargs={"mailing_list_slug": mailing_list.slug}
            ),
            {
                "email": "test@test.test",
                "first_name": "Testy",
                "last_name": "McTestface",
                "are_you_sure": True,
            },
        )
        assert models.Subscription.objects.all().exists()
        subscription = models.Subscription.objects.first()
        assert subscription.mailing_list == mailing_list
        assert subscription.user.email == "test@test.test"
        assert subscription.user.first_name == "Testy"
        assert subscription.user.last_name == "McTestface"

    def test_post_first_name_only(self, client, mailing_list):
        assert not models.Subscription.objects.all().exists()
        client.post(
            reverse(
                "mailinglist:subscribe", kwargs={"mailing_list_slug": mailing_list.slug}
            ),
            {
                "email": "test@test.test",
                "first_name": "Testy",
                "last_name": "",
                "are_you_sure": True,
            },
        )
        assert models.Subscription.objects.all().exists()
        subscription = models.Subscription.objects.first()
        assert subscription.user.email == "test@test.test"
        assert subscription.user.first_name == "Testy"
        assert subscription.user.last_name == ""

    def test_post_bad_email(self, client, mailing_list):
        assert not models.Subscription.objects.all().exists()
        response = client.post(
            reverse(
                "mailinglist:subscribe", kwargs={"mailing_list_slug": mailing_list.slug}
            ),
            {
                "email": "notanemailaddress.com",
                "first_name": "Testy",
                "last_name": "",
                "are_you_sure": True,
            },
        )
        assert not models.Subscription.objects.all().exists()
        assert not response.context["form"].is_valid()

    def test_post_not_sure(self, client, mailing_list):
        assert not models.Subscription.objects.all().exists()
        response = client.post(
            reverse(
                "mailinglist:subscribe", kwargs={"mailing_list_slug": mailing_list.slug}
            ),
            {
                "email": "test@test.test",
                "first_name": "Testy",
                "last_name": "McTestface",
                "are_you_sure": False,
            },
        )
        assert not models.Subscription.objects.all().exists()
        assert not response.context["form"].is_valid()


class TestSubscribeConfirmView:
    def test_good_subscribe(self, client, mailing_list, subscription):
        client.get(
            reverse(
                "mailinglist:subscribe_confirm", kwargs={"token": subscription.token}
            )
        )
        subscription.refresh_from_db()
        assert subscription.status == enum.SubscriptionStatusEnum.SUBSCRIBED

    def test_good_subscribe_global_unsubscribe(
        self, client, mailing_list, subscription
    ):
        subscription.mailing_list = None
        subscription.save()
        client.get(
            reverse(
                "mailinglist:subscribe_confirm", kwargs={"token": subscription.token}
            )
        )
        subscription.refresh_from_db()
        assert hasattr(subscription.user, "mailinglist_deny")
        assert subscription.status == enum.SubscriptionStatusEnum.SUBSCRIBED

    def test_bad_subscribe(self, client, mailing_list, subscription):
        response = client.get(
            reverse(
                "mailinglist:subscribe_confirm",
                kwargs={"token": subscription.token + "a"},
            )
        )
        subscription.refresh_from_db()
        assert subscription.status == enum.SubscriptionStatusEnum.PENDING
        assert b"token in the URL is invalid! Sorry." in response.content


class TestUnsubscribeView:
    def test_good_unsubscribe(self, client, mailing_list, active_subscription):
        client.get(
            reverse(
                "mailinglist:unsubscribe", kwargs={"token": active_subscription.token}
            )
        )
        active_subscription.refresh_from_db()
        assert active_subscription.status == enum.SubscriptionStatusEnum.UNSUBSCRIBED

    def test_bad_unsubscribe(self, client, mailing_list, active_subscription):
        response = client.get(
            reverse(
                "mailinglist:subscribe_confirm",
                kwargs={"token": active_subscription.token + "a"},
            )
        )
        active_subscription.refresh_from_db()
        assert active_subscription.status == enum.SubscriptionStatusEnum.SUBSCRIBED
        assert b"token in the URL is invalid! Sorry." in response.content


class TestArchiveIndexView:
    def test_archive_visibility(self, client, mailing_list):
        # should be visible by default
        response = client.get(
            reverse(
                "mailinglist:archive_index",
                kwargs={"mailing_list_slug": mailing_list.slug},
            )
        )
        assert b"Archive of test list" in response.content

    def test_archive_invisibility(self, client, mailing_list):
        mailing_list.visible = False
        mailing_list.save()
        response = client.get(
            reverse(
                "mailinglist:archive_index",
                kwargs={"mailing_list_slug": mailing_list.slug},
            )
        )
        assert response.status_code == 404

    def test_archive_message_invisibility(
        self, client, submission, message, mailing_list
    ):
        # should be visible by default
        response = client.get(
            reverse(
                "mailinglist:archive_index",
                kwargs={"mailing_list_slug": mailing_list.slug},
            )
        )
        assert b"test message" not in response.content

    def test_archive_message_visibility(
        self, client, published_submission, message, mailing_list
    ):
        response = client.get(
            reverse(
                "mailinglist:archive_index",
                kwargs={"mailing_list_slug": mailing_list.slug},
            )
        )
        assert b"test message" in response.content


class TestArchiveView:
    def test_message_visibility(
        self, client, published_submission, message, mailing_list
    ):
        response = client.get(
            reverse(
                "mailinglist:archive",
                kwargs={
                    "mailing_list_slug": mailing_list.slug,
                    "message_slug": message.slug,
                },
            )
        )
        assert response.status_code == 200
        assert b"test message" in response.content

    def test_message_invisibility_mailinglist(
        self, client, published_submission, message, mailing_list
    ):
        mailing_list.visible = False
        mailing_list.save()
        response = client.get(
            reverse(
                "mailinglist:archive",
                kwargs={
                    "mailing_list_slug": mailing_list.slug,
                    "message_slug": message.slug,
                },
            )
        )
        assert response.status_code == 404

    def test_message_invisibility_unpublished(
        self, client, submission, message, mailing_list
    ):
        response = client.get(
            reverse(
                "mailinglist:archive",
                kwargs={
                    "mailing_list_slug": mailing_list.slug,
                    "message_slug": message.slug,
                },
            )
        )
        assert response.status_code == 404



# Generated by Django 4.2.7 on 2023-12-03 14:41

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("mailinglist", "0002_messageattachment"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="messagepart",
            options={"ordering": ["order"]},
        ),
    ]



from django import forms
from django.db.models import Q

from mailinglist.addressimport.parsers import parse_csv
from mailinglist.models import MailingList, Message, Submission


class ImportForm(forms.Form):
    def clean(self):
        # If there are validation errors earlier on, don't bother.
        if not (
            "address_file" in self.cleaned_data
            and "ignore_errors" in self.cleaned_data
            and "mailing_list" in self.cleaned_data
        ):
            return self.cleaned_data

        address_file = self.cleaned_data["address_file"]

        content_type = address_file.content_type
        allowed_types = (
            "text/plain",
            "text/comma-separated-values",
            "text/csv",
            "application/csv",
        )
        if content_type not in allowed_types:
            raise forms.ValidationError(
                f"File type '{content_type}' was not recognized."
            )

        ext = address_file.name.rsplit(".", 1)[-1].lower()
        parser = None
        if ext == "csv":
            parser = parse_csv
        if parser is None:
            raise forms.ValidationError(f"File extension '{ext}' was not recognized.")

        self.addresses = parser(
            address_file.file,
            self.cleaned_data["mailing_list"],
            self.cleaned_data["ignore_errors"],
        )

        if len(self.addresses) == 0:
            raise forms.ValidationError("No entries could found in this file.")

        return self.cleaned_data

    def get_addresses(self):
        return getattr(self, "addresses", {})

    mailing_list = forms.ModelChoiceField(
        label="Mailing List",
        queryset=MailingList.objects.all(),
    )
    address_file = forms.FileField(label="Address file")
    ignore_errors = forms.BooleanField(
        label="Ignore non-fatal errors", initial=True, required=False
    )


class ConfirmForm(forms.Form):
    def clean(self):
        value = self.cleaned_data.get("confirm", False)

        if not value:
            raise forms.ValidationError("You should confirm in order to continue.")
        return self.cleaned_data

    confirm = forms.BooleanField(
        label="Confirm import", initial=True, widget=forms.HiddenInput
    )


class SubmissionModelForm(forms.ModelForm):
    class Meta:
        model = Submission
        fields = "__all__"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        submission = kwargs.get("instance")
        if submission is not None:
            self.fields[
                "exclude"
            ].queryset = submission.message.mailing_list.subscriptions.all()
            self.fields["message"].queryset = Message.objects.filter(
                Q(submission__isnull=True) | Q(submission=submission)
            )
        else:
            self.fields["message"].queryset = Message.objects.filter(
                submission__isnull=True
            )



from django_enumfield.enum import Enum


class SubscriptionStatusEnum(Enum):
    PENDING = 0
    SUBSCRIBED = 1
    UNSUBSCRIBED = 2

    __default__ = PENDING


class SubmissionStatusEnum(Enum):
    NEW = 0
    PENDING = 1
    SENDING = 2
    SENT = 3

    __default__ = NEW

    __transitions__ = {
        SENT: (SENDING,),
        SENDING: (PENDING,),
        PENDING: (NEW,),
    }



from mailinglist.services import SubmissionService

try:
    from celery import shared_task
except ImportError:
    pass
else:

    @shared_task
    def process_submissions():
        SubmissionService().process_submissions()



import os

from django.apps import apps
from django.conf import settings
from django.core.exceptions import ValidationError
from django.core.mail import EmailMultiAlternatives


class MailinglistDefaultHookset:
    def create_user(self, *, email, first_name, last_name):
        user_model = apps.get_model(settings.MAILINGLIST_USER_MODEL)
        try:
            user = user_model.objects.get(email=email)
        except user_model.DoesNotExist:
            user = user_model.objects.create(
                username=f"user{hash(email)}",
                email=email,
                first_name=first_name,
                last_name=last_name,
            )
        return user

    def send_message(
        self,
        *,
        to,
        body,
        subject,
        from_email,
        html_body=None,
        attachments=None,
        headers=None,
    ):
        message = EmailMultiAlternatives(
            to=to,
            body=body,
            subject=subject,
            from_email=from_email,
            headers=headers,
        )
        _attachments = attachments or []

        for attachment in _attachments:
            message.attach_file(attachment.file.path)

        if html_body is not None:
            message.attach_alternative(html_body, "text/html")
        message.send()

    def message_attachment_file_validator(self, value):
        valid_file_extensions = [".pdf", ".jpg", ".jpeg", ".png", ".tiff", ".tif"]
        ext = os.path.splitext(value.name)[-1]
        if ext.lower() not in valid_file_extensions:
            raise ValidationError(
                f"File must have "
                f"{' or '.join([f'{e!r}' for e in valid_file_extensions]) } "
                "as its extension"
            )



from django.urls import path
from django.views.generic import TemplateView

from mailinglist import views

app_name = "mailinglist"

urlpatterns = [
    path(
        "global_deny/",
        views.GlobalDenyView.as_view(),
        name="global_deny",
    ),
    path(
        "global_deny/success/",
        TemplateView.as_view(
            template_name="mailinglist/web/global_unsubscribe_success.html"
        ),
        name="global_deny_success",
    ),
    path(
        "subscribe/<slug:mailing_list_slug>/",
        views.SubscribeView.as_view(),
        name="subscribe",
    ),
    path(
        "subscribe/<slug:mailing_list_slug>/success/",
        views.SubscribeSuccessView.as_view(),
        name="subscribe_success",
    ),
    path(
        "subscribe/<slug:token>/confirm/",
        views.SubscribeConfirmView.as_view(),
        name="subscribe_confirm",
    ),
    path(
        "subscriptions/<slug:token>/",
        views.SubscriptionView.as_view(),
        name="subscriptions",
    ),
    path(
        "unsubscribe/<slug:token>/",
        views.UnsubscribeView.as_view(),
        name="unsubscribe",
    ),
    path(
        "archive/",
        views.ArchivesView.as_view(),
        name="archives",
    ),
    path(
        "archive/<slug:mailing_list_slug>/",
        views.ArchiveIndexView.as_view(),
        name="archive_index",
    ),
    path(
        "archive/<slug:mailing_list_slug>/<slug:message_slug>/",
        views.ArchiveView.as_view(),
        name="archive",
    ),
]



import sys
import builtins
from unittest.mock import patch
from importlib import reload

import pytest
from django.core.management import call_command


@patch("mailinglist.services.SubmissionService.process_submissions")
def test_process_submissions_managment_command(p_process):
    call_command("process_submissions")
    p_process.assert_called_once_with()


@pytest.fixture
def hide_celery(monkeypatch):
    import_orig = builtins.__import__

    def mocked_import(name, *args, **kwargs):
        if name == "celery":
            raise ImportError()
        return import_orig(name, *args, **kwargs)

    monkeypatch.setattr(builtins, "__import__", mocked_import)


@pytest.mark.usefixtures("hide_celery")
def test_process_submissions_no_celery():
    from mailinglist import tasks

    assert not hasattr(tasks, "process_submissions")


@patch("mailinglist.services.SubmissionService.process_submissions")
def test_process_submissions_celery(p_process):
    reload(sys.modules["mailinglist.tasks"])
    from mailinglist import tasks

    assert hasattr(tasks, "process_submissions")
    tasks.process_submissions()
    p_process.assert_called_once_with()



from django.http import Http404
from django.urls import reverse
from django.utils.timezone import now
from django.views.generic import DetailView, FormView, ListView, TemplateView
from django.views.generic.detail import (
    SingleObjectMixin,
    SingleObjectTemplateResponseMixin,
)

from mailinglist import models
from mailinglist.conf import settings
from mailinglist.forms import SubscribeForm, SubscriptionForm
from mailinglist.services import SubscriptionService


class DetailFormView(
    SingleObjectTemplateResponseMixin,
    SingleObjectMixin,
    FormView,
):
    def get(self, request, *args, **kwargs):
        self.object = self.get_object()
        return super().get(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        return super().post(request, *args, **kwargs)


class SubscribeFormMixin:
    form_class = SubscribeForm

    def get_initial(self):
        initial = super().get_initial()
        if not self.request.user.is_anonymous:
            initial.update(
                {
                    "email": self.request.user.email,
                    "first_name": self.request.user.first_name,
                    "last_name": self.request.user.last_name,
                }
            )
        return initial

    def form_valid(self, form):
        service = SubscriptionService()
        user = service.create_user(
            email=form.cleaned_data["email"],
            first_name=form.cleaned_data["first_name"],
            last_name=form.cleaned_data["last_name"],
        )
        service.subscribe(user=user, mailing_list=self.get_object())
        return super().form_valid(form)


class SubscribeView(SubscribeFormMixin, DetailFormView):
    """Allows users to subscrbe to a mailing list."""

    template_name = "mailinglist/web/subscribe.html"
    queryset = models.MailingList.objects.all()
    slug_url_kwarg = "mailing_list_slug"

    def get_success_url(self):
        return reverse("mailinglist:subscribe_success", kwargs=self.kwargs)


class GlobalDenyView(SubscribeFormMixin, FormView):
    """Allows users to subscrbe to the global deny list."""

    template_name = "mailinglist/web/global_unsubscribe.html"

    def get_success_url(self):
        return reverse("mailinglist:global_deny_success")

    def get_object(self):
        return None


class SubscribeSuccessView(DetailView):
    """Provides user with insight about subscription, namely whether or
    not they will need to confirm their email address"""

    template_name = "mailinglist/web/subscribe_success.html"
    queryset = models.MailingList.objects.all()
    slug_url_kwarg = "mailing_list_slug"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update(
            {
                "requires_confirmation": settings.MAILINGLIST_CONFIRM_EMAIL_SUBSCRIBE,
            }
        )
        return context


class SubscriptionView(DetailFormView):
    """Allows the user to view what mailing lists they are subscribed to."""

    # Since the subscription might not point to a user who can log in to the
    #  site, we must not rely on `request.user`, we will use the subscription token
    #  as a bearer token and allow the other subscriptions that the
    #  `subscription.user` has.
    template_name = "mailinglist/web/subscriptions.html"
    form_class = SubscriptionForm
    queryset = models.Subscription.objects.all()
    slug_url_kwarg = "token"
    slug_field = "token"

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        instance = self.get_object()
        kwargs.update({"user": instance.user})
        return kwargs

    def form_valid(self, form):
        form.save()
        return super().form_valid(form)

    def get_success_url(self):
        return reverse("mailinglist:subscriptions", kwargs=self.kwargs)


class IsSubscriptionMixin:
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        _is_subcription = self.subscription is not None
        _is_global_unsubcription = False
        if _is_subcription and self.subscription.mailing_list is None:
            _is_global_unsubcription = True
        context.update(
            {
                "token": self.kwargs.get("token"),
                "is_global_unsubscription": _is_global_unsubcription,
                "is_subscription": _is_subcription,
            }
        )
        return context


class SubscribeConfirmView(IsSubscriptionMixin, TemplateView):
    """Reachable when user clicks on link in confirmation message, tells
    user that they are fully subscribed."""

    template_name = "mailinglist/web/subscribe_confirm.html"

    # Do not want _any_ subscription data leaking into the UI, so keep
    #   it a simple template view
    def get(self, request, *args, **kwargs):
        self.subscription = SubscriptionService().confirm_subscription(
            token=kwargs.get("token")
        )
        return super().get(request, *args, **kwargs)


class UnsubscribeView(IsSubscriptionMixin, TemplateView):
    """Reachable when user clicks unsubscribe link in any sent message.
    Deactivates user's subscription to the mailing list and provides link
    to user's subscriptions page"""

    # Do not want _any_ subscription data leaking into the UI, so keep
    #   it a simple template view
    template_name = "mailinglist/web/unsubscribe.html"

    def get(self, request, *args, **kwargs):
        self.subscription = SubscriptionService().unsubscribe(token=kwargs.get("token"))
        return super().get(request, *args, **kwargs)


class ArchivesView(ListView):
    """Allows user to browse messages published on a mailing list."""

    template_name = "mailinglist/web/archive/archives.html"
    queryset = models.MailingList.objects.filter(visible=True)


class ArchiveIndexView(DetailView):
    """Allows user to browse messages published on a mailing list."""

    template_name = "mailinglist/web/archive/index.html"
    queryset = models.MailingList.objects.filter(visible=True)
    slug_url_kwarg = "mailing_list_slug"


class ArchiveView(DetailView):
    """Allows user to view messages published on a mailing list."""

    template_name = "mailinglist/web/archive/message.html"
    queryset = models.Message.objects.filter(mailing_list__visible=True)

    def get_object(self, **kwargs):
        qs = super().get_queryset(**kwargs)
        try:
            return qs.get(
                slug=self.kwargs.get("message_slug"),
                mailing_list__slug=self.kwargs.get("mailing_list_slug"),
                submission__published__lte=now(),
            )
        except models.Message.DoesNotExist:
            raise Http404("No message like that")



from unittest.mock import patch, call
from django.core.exceptions import ValidationError
import pytest
import io

from mailinglist.addressimport.parsers import AddressList, parse_csv


@pytest.fixture
def address_list(mailing_list):
    return AddressList(mailing_list=mailing_list)


@pytest.fixture
def address_list_quiet(mailing_list):
    return AddressList(mailing_list=mailing_list, ignore_errors=True)


class TestAddressList:
    def test_add(self, address_list):
        address_list.add(email="test@good.test", first_name="Test", last_name="Good")
        assert "test@good.test" in address_list.addresses
        assert address_list.addresses["test@good.test"] == {
            "email": "test@good.test",
            "first_name": "Test",
            "last_name": "Good",
        }

    def test_add_bad_email_quiet(self, address_list_quiet):
        address_list_quiet.add(email="bad_email", first_name="Test", last_name="Good")
        assert "bad_email" not in address_list_quiet.addresses

    def test_add_bad_email(self, address_list):
        with pytest.raises(ValidationError):
            address_list.add(email="bad_email", first_name="Test", last_name="Good")
        assert "bad_email" not in address_list.addresses

    def test_add_email_overlong_quiet(self, address_list_quiet):
        address_list_quiet.add(
            email="long@email.com" + ("m" * 300), first_name="Test", last_name="Good"
        )
        assert "long@email.com" not in address_list_quiet.addresses

    def test_add_email_overlong(self, address_list):
        with pytest.raises(ValidationError):
            address_list.add(
                email="long@email.com" + ("m" * 300),
                first_name="Test",
                last_name="Good",
            )
        assert not address_list.addresses

    def test_add_first_name_overlong_quiet(self, address_list_quiet):
        address_list_quiet.add(
            email="long@email.com",
            first_name="Test" + ("m" * 300),
            last_name="Good",
        )
        assert "long@email.com" in address_list_quiet.addresses
        assert len(address_list_quiet.addresses["long@email.com"]["first_name"]) < 300

    def test_add_first_name_overlong(self, address_list):
        with pytest.raises(ValidationError):
            address_list.add(
                email="long@email.com",
                first_name="Test" + ("m" * 300),
                last_name="Good",
            )

    def test_add_last_name_overlong_quiet(self, address_list_quiet):
        address_list_quiet.add(
            email="long@email.com",
            first_name="Test",
            last_name="Good" + ("m" * 300),
        )
        assert "long@email.com" in address_list_quiet.addresses
        assert len(address_list_quiet.addresses["long@email.com"]["last_name"]) < 300

    def test_add_last_name_overlong(self, address_list):
        with pytest.raises(ValidationError):
            address_list.add(
                email="long@email.com",
                first_name="Test",
                last_name="Good" + ("m" * 300),
            )
        assert not address_list.addresses

    def test_add_no_last_name(self, address_list):
        address_list.add(
            email="good@email.com",
            first_name="Test",
            last_name=None,
        )
        assert "good@email.com" in address_list.addresses
        assert address_list.addresses["good@email.com"] == {
            "email": "good@email.com",
            "first_name": "Test",
            "last_name": None,
        }

    def test_add_duplicate_quiet(self, address_list_quiet):
        address_list_quiet.add(
            email="good@email.com",
            first_name="Test",
            last_name=None,
        )
        address_list_quiet.add(
            email="good@email.com",
            first_name="Other",
            last_name="Test",
        )
        assert "good@email.com" in address_list_quiet.addresses
        assert address_list_quiet.addresses["good@email.com"] == {
            "email": "good@email.com",
            "first_name": "Test",
            "last_name": None,
        }

    def test_add_duplicate(self, address_list):
        address_list.add(
            email="good@email.com",
            first_name="Test",
            last_name=None,
        )
        with pytest.raises(ValidationError):
            address_list.add(
                email="good@email.com",
                first_name="Other",
                last_name="Test",
            )

    def test_add_duplicate_subscription(self, address_list, subscription):
        with pytest.raises(ValidationError):
            address_list.add(
                email=subscription.user.email,
                first_name="Test",
                last_name=None,
            )

    def test_add_duplicate_subscription_quiet(self, address_list_quiet, subscription):
        address_list_quiet.add(
            email=subscription.user.email,
            first_name="Test",
            last_name=None,
        )
        assert subscription.user.email not in address_list_quiet.addresses


class TestParseCsv:
    @patch.object(AddressList, "add")
    def test_parse(self, p_add, mailing_list):
        _file = io.BytesIO(
            b"""first_name,last_name,email
test,person,person@person.us
another,human,identity@real.yes"""
        )
        parse_csv(_file, mailing_list)
        p_add.assert_has_calls(
            [
                call(
                    first_name="test",
                    last_name="person",
                    email="person@person.us",
                    location="line 0",
                ),
                call(
                    first_name="another",
                    last_name="human",
                    email="identity@real.yes",
                    location="line 1",
                ),
            ]
        )



import importlib

from appconf import AppConf
from appconf.utils import import_attribute
from django.conf import settings  # noqa: F401
from django.core.exceptions import ImproperlyConfigured


class MailinglistAppConf(AppConf):
    BASE_URL = None
    DEFAULT_SENDER_EMAIL = None
    DEFAULT_SENDER_NAME = "Administrator"
    USER_MODEL = settings.AUTH_USER_MODEL
    HOOKSET = "mailinglist.hooks.MailinglistDefaultHookset"
    CONFIRM_EMAIL_SUBSCRIBE = True
    EMAIL_DELAY = 0.1
    BATCH_DELAY = 10  # seconds
    BATCH_SIZE = 100

    def configure_hookset(self, value):
        return import_attribute(value)()

    def configure_default_sender_email(self, value):
        if value is None:
            raise ImproperlyConfigured(
                "Must configure MAILINGLIST_DEFAULT_SENDER_EMAIL"
            )
        return value

    def configure_base_url(self, value):
        if value is None:
            raise ImproperlyConfigured("Must configure MAILINGLIST_BASE_URL")
        return value


class HookProxy:
    def __getattr__(self, attr):
        return getattr(settings.MAILINGLIST_HOOKSET, attr)


hookset = HookProxy()



from django import forms

from mailinglist import models
from mailinglist.enum import SubscriptionStatusEnum
from mailinglist.services import SubscriptionService


class SubscribeForm(forms.Form):
    """Basic subscription details."""

    email = forms.EmailField(label="Email:")
    first_name = forms.CharField(label="First name:")
    last_name = forms.CharField(label="Last name:", required=False)
    are_you_sure = forms.BooleanField(label="Are you sure?")


class SubscriptionForm(forms.Form):
    """Dynamically generates a series of checkboxes for each visible
    mailing list so that users can subscribe at their will. Also includes
    checkbox for global unsubscribe."""

    def __init__(self, *args, user, **kwargs):
        super().__init__(*args, **kwargs)
        subscriptions = user.subscriptions.filter(
            status=SubscriptionStatusEnum.SUBSCRIBED
        )
        self.user = user
        self.subscribed_lists = {
            s.mailing_list.pk: s for s in subscriptions if s.mailing_list
        }
        for mailing_list in models.MailingList.objects.filter(visible=True):
            self.fields[f"mailing-list_{mailing_list.pk}"] = forms.BooleanField(
                required=False,
                initial=mailing_list.pk in self.subscribed_lists,
                label=mailing_list.name,
            )
        self.fields["global-deny"] = forms.BooleanField(
            required=False,
            initial=hasattr(user, "mailinglist_deny"),
            label="Block all mailinglists from sending to me!",
        )

    def save(self, commit=True):
        if not commit:
            return
        if self.cleaned_data.get("global-deny", False):
            models.GlobalDeny.objects.get_or_create(user=self.user)
        else:
            models.GlobalDeny.objects.filter(user=self.user).delete()
        service = SubscriptionService()
        for mailing_list in models.MailingList.objects.filter(visible=True):
            wants = self.cleaned_data.get(f"mailing-list_{mailing_list.pk}", False)
            if wants and mailing_list.pk not in self.subscribed_lists:
                # create subscription
                service.subscribe(
                    user=self.user, mailing_list=mailing_list, force_confirm=True
                )
            if not wants and mailing_list.pk in self.subscribed_lists:
                service._confirm_unsubscription(self.subscribed_lists[mailing_list.pk])



from django.core.management.base import BaseCommand

from mailinglist.services import SubmissionService


class Command(BaseCommand):
    help = "Send published submissions."

    def handle(self, *args, **options):
        SubmissionService().process_submissions()



import time
from random import randint

from django.conf import settings
from django.template.loader import select_template
from django.urls import reverse
from django.utils.crypto import get_random_string
from django.utils.timezone import now

from mailinglist import models
from mailinglist.conf import hookset
from mailinglist.enum import SubmissionStatusEnum, SubscriptionStatusEnum


class TemplateSet:
    """Represents the templates needed for generating outgoing email."""

    def __init__(
        self, *, mailing_list: models.MailingList = None, action: str = "message"
    ):
        self.mailing_list = mailing_list
        self.action = action
        self._templates = None

    @property
    def templates(self):
        if self._templates is not None:
            return self._templates
        self._templates = {}
        self._templates["subject"] = self._get_template(_for=f"{self.action}_subject")
        self._templates["body"] = self._get_template(_for=self.action)
        if self.mailing_list is None or self.mailing_list.send_html:
            self._templates["html_body"] = self._get_template(
                _for=self.action, suffix="html"
            )
        return self._templates

    def _get_default_context(self):
        return {
            "mailing_list": self.mailing_list,
            "BASE_URL": settings.MAILINGLIST_BASE_URL,
            "DEFAULT_SENDER_NAME": settings.MAILINGLIST_DEFAULT_SENDER_NAME,
        }

    def _get_template(self, *, suffix="txt", _for="message"):
        _root = "mailinglist/email"
        slug = "global-deny"
        if self.mailing_list is not None:
            slug = self.mailing_list.slug
        return select_template(
            [
                f"{_root}/{slug}/{_for}.{suffix}",
                f"{_root}/{_for}.{suffix}",
            ]
        )

    def render_to_dict(self, context: dict):  # -> dict[str, str]:
        """Renders each template in the set and arranges the text into a
        dictionary, requires the context for the specific message and
        subscription"""

        rendered = {}
        _context = self._get_default_context()
        _context.update(context)
        for attr, template in self.templates.items():
            rendered[attr] = template.render(_context).strip()
        return rendered


class MessageService:
    """Composes email for sending"""

    def _urlify(self, path):
        return f"<{settings.MAILINGLIST_BASE_URL}{path}>"

    def _mailto(self, mailing_list, subject=None, appended=True):
        if mailing_list is None:
            return ""
        if subject is None:
            subject = mailing_list.slug
        buffer = ""
        if appended:
            buffer += ","

        buffer += f"<mailto: {mailing_list.email}?subject={subject}>"
        return buffer

    def _headers(self, *, subscription):
        """Provides necessary headers with correct formatting for good
        adherence to RFC2369.

        Reference
        ^^^^^^^^^
        https://datatracker.ietf.org/doc/html/rfc2369
        """
        _list = subscription.mailing_list
        _help_path = reverse(
            "mailinglist:subscriptions", kwargs={"token": subscription.token}
        )
        _unsubscribe_path = reverse(
            "mailinglist:unsubscribe", kwargs={"token": subscription.token}
        )
        _subscribe_path = reverse(
            "mailinglist:subscribe_confirm", kwargs={"token": subscription.token}
        )
        if subscription.mailing_list is None:
            _archive_path = reverse("mailinglist:archives")
        else:
            _archive_path = reverse(
                "mailinglist:archive_index", kwargs={"mailing_list_slug": _list.slug}
            )
        return {
            "List-Help": self._urlify(_help_path) + self._mailto(_list, subject="help"),
            "List-Unsubscribe": self._urlify(_unsubscribe_path)
            + self._mailto(_list, subject="unsubscribe"),
            "List-Subscribe": self._urlify(_subscribe_path),
            "List-Post": "NO",
            "List-Owner": self._mailto(_list, appended=False),
            "List-Archive": self._urlify(_archive_path),
        }

    def _prepare_kwargs(
        self,
        *,
        subscription: models.Subscription,
        template_set: TemplateSet,
        message: models.Message = None,
    ):  # -> dict:
        _kwargs = template_set.render_to_dict(
            context={"subscription": subscription, "message": message}
        )
        _kwargs.update(
            {
                "to": [subscription.user.email],
                "headers": self._headers(subscription=subscription),
            }
        )
        return _kwargs

    def prepare_message_kwargs(
        self,
        *,
        subscription: models.Subscription,
        template_set: TemplateSet,
        message: models.Message,
    ):  # -> dict:
        """Composes and structures outgoing message data for email sending"""
        return self._prepare_kwargs(
            message=message,
            subscription=subscription,
            template_set=template_set,
        )

    def prepare_confirmation_kwargs(
        self, *, subscription: models.Subscription, template_set: TemplateSet
    ):
        """Composes and structures outgoing message data for confirmation email"""
        return self._prepare_kwargs(
            subscription=subscription,
            template_set=template_set,
        )

    def _prepare_preview(self, *, message):
        template_set = TemplateSet(mailing_list=message.mailing_list)
        rendered = template_set.render_to_dict(
            context={"subscription": None, "message": message}
        )
        return rendered

    def prepare_message_preview(self, *, message):
        return self._prepare_preview(message=message)["body"]

    def prepare_message_preview_html(self, *, message):
        return self._prepare_preview(message=message).get("html_body", None)


class SubscriptionService:
    """Manages all subscription and unsubscribe events."""

    def _random_string(self, length):
        return get_random_string(
            length=length, allowed_chars="abcdefghijklmnopqrstuvwxyz0123456789-"
        )

    def _generate_token(self, *, user, mailing_list):
        _s = randint(0, len(user.email) - 1)
        _hash = str(hash(f"{user.email[0:_s]}{mailing_list}{user.email[_s:0]}"))
        _prefix = self._random_string(length=45 - len(_hash))
        return _prefix + _hash

    def _rotate_token(self, subscription):
        subscription.token = self._generate_token(
            user=subscription.user, mailing_list=subscription.mailing_list
        )
        subscription.save()

    def _new_subscription(self, *, user, mailing_list):
        token = self._generate_token(user=user, mailing_list=mailing_list)
        subscription = models.Subscription.objects.create(
            user=user,
            mailing_list=mailing_list,
            token=token,
        )
        return subscription

    def _update_subscription_status(self, *, subscription, to_status):
        if subscription.status == to_status:
            return subscription
        models.SubscriptionChange.objects.create(
            subscription=subscription,
            from_status=subscription.status,
            to_status=to_status,
        )
        subscription.status = to_status
        subscription.save()
        return subscription

    def _confirm_subscription(self, subscription):
        if subscription.mailing_list is None:
            models.GlobalDeny.objects.get_or_create(user=subscription.user)
        return self._update_subscription_status(
            subscription=subscription, to_status=SubscriptionStatusEnum.SUBSCRIBED
        )

    def _send_subscription_confirmation(self, subscription):
        sender = (
            f'"{settings.MAILINGLIST_DEFAULT_SENDER_NAME}" '
            f"<{settings.MAILINGLIST_DEFAULT_SENDER_EMAIL}>"
        )
        if subscription.mailing_list is not None:
            sender = subscription.mailing_list.sender_tag
        hookset.send_message(
            from_email=sender,
            **MessageService().prepare_confirmation_kwargs(
                subscription=subscription,
                template_set=TemplateSet(
                    mailing_list=subscription.mailing_list,
                    action="subscribe",
                ),
            ),
        )

    # Type hints get bothersome for this dynamic user model...
    def create_user(self, *, email: str, first_name: str, last_name: str):
        """Creates a "user" for a new subscription. This method calls
        the same-named method in the hookset to actually perform the action."""
        user = hookset.create_user(
            email=email, first_name=first_name, last_name=last_name
        )
        return user

    def _subscribe(self, *, user, mailing_list):
        try:
            subscription = models.Subscription.objects.get(
                user=user, mailing_list=mailing_list
            )
        except models.Subscription.DoesNotExist:
            subscription = self._new_subscription(user=user, mailing_list=mailing_list)
        return subscription

    def force_subscribe(
        self, *, user, mailing_list: models.MailingList
    ):  # -> models.Subscription:
        """Creates an active subscription skipping any confirmation email."""
        subscription = self._subscribe(user=user, mailing_list=mailing_list)
        self._confirm_subscription(subscription)
        return subscription

    def subscribe(
        self, *, user, mailing_list: models.MailingList, force_confirm=False
    ):  # -> models.Subscription:
        """Creates a subscription and sends the activation email (or just
        activates it based on settings)"""
        if models.GlobalDeny.objects.filter(user=user).exists():
            return
        subscription = self._subscribe(user=user, mailing_list=mailing_list)
        if subscription.status == SubscriptionStatusEnum.SUBSCRIBED:
            return subscription

        if mailing_list is None or (
            settings.MAILINGLIST_CONFIRM_EMAIL_SUBSCRIBE and not force_confirm
        ):
            self._send_subscription_confirmation(subscription)
        else:
            self._confirm_subscription(subscription)
        return subscription

    def confirm_subscription(self, *, token: str):  # -> models.Subscription:
        """Activates a subscription"""
        # get subscription
        try:
            subscription = models.Subscription.objects.get(token=token)
        except models.Subscription.DoesNotExist:
            # nothing to see here
            return None
        return self._confirm_subscription(subscription)

    def _confirm_unsubscription(self, subscription):
        return self._update_subscription_status(
            subscription=subscription, to_status=SubscriptionStatusEnum.UNSUBSCRIBED
        )

    def unsubscribe(self, *, token: str):  # -> models.Subscription:
        """Deactivates a subscription"""
        try:
            subscription = models.Subscription.objects.get(token=token)
        except models.Subscription.DoesNotExist:
            # nothing to see here
            return
        return self._confirm_unsubscription(subscription)


class SubmissionService:
    """Manages send activities for published submissions."""

    def _get_included_subscribers(self, submission):
        # get current list of subscribers
        subscriptions = (
            submission.message.mailing_list.subscriptions.filter(
                status=SubscriptionStatusEnum.SUBSCRIBED
            )
            # remove all global denies
            .filter(user__mailinglist_deny__isnull=True)
            # remove all excludes
            .exclude(pk__in=submission.exclude.all().values_list("id", flat=True))
        )
        return subscriptions

    def _send_message(self, *, message, subscription, template_set, **kwargs):
        hookset.send_message(
            from_email=subscription.mailing_list.sender_tag,
            **MessageService().prepare_message_kwargs(
                message=message, subscription=subscription, template_set=template_set
            ),
            **kwargs,
        )

    def _ensure_sent(self, *, subscription, submission, **kwargs):
        """Idempotent sending of message, returns ``True`` if email was
        actually sent (returns ``False`` if email was sent previously)."""
        # check if this has been sent already
        sending_kwargs = {
            "submission": submission,
            "subscription": subscription,
        }
        if models.Sending.objects.filter(**sending_kwargs).exists():
            return False
        # send email
        self._send_message(
            message=submission.message, subscription=subscription, **kwargs
        )
        # log sending of email
        models.Sending.objects.create(**sending_kwargs)
        return True

    def _rate_limit(self, total_send_count):
        batch_sleep = False
        if settings.MAILINGLIST_BATCH_DELAY is not None:
            if total_send_count % settings.MAILINGLIST_BATCH_SIZE == 0:
                batch_sleep = True
                time.sleep(settings.MAILINGLIST_BATCH_DELAY)
        if not batch_sleep and settings.MAILINGLIST_EMAIL_DELAY is not None:
            time.sleep(settings.MAILINGLIST_EMAIL_DELAY)

    def process_submission(
        self, submission: models.Submission, *, send_count: int = 0
    ):  # -> int:
        """Sends submitted message to each (non-excluded) subscriber,
        observing rate limits configured in settings."""
        submission.status = SubmissionStatusEnum.SENDING
        submission.save()
        subscriptions = self._get_included_subscribers(submission)
        template_set = TemplateSet(mailing_list=submission.message.mailing_list)
        attachments = list(submission.message.attachments.all())
        for subscription in subscriptions:
            did_send = self._ensure_sent(
                submission=submission,
                subscription=subscription,
                template_set=template_set,
                attachments=attachments,
            )
            if not did_send:
                continue
            send_count += 1
            self._rate_limit(send_count)
        submission.status = SubmissionStatusEnum.SENT
        submission.save()
        return send_count

    def _get_outstanding_submissions(self):
        sending_submissions = models.Submission.objects.filter(
            status=SubmissionStatusEnum.SENDING
        )
        published_submissions = models.Submission.objects.filter(
            status=SubmissionStatusEnum.PENDING,
            published__isnull=False,
            published__lte=now(),
        )
        return list(sending_submissions) + list(published_submissions)

    def process_submissions(self):  # -> None:
        """Finds all unsent published ``Submission`` instances and sends them."""
        send_count = 0
        for published_submission in self._get_outstanding_submissions():
            send_count = self.process_submission(
                published_submission, send_count=send_count
            )

    def publish(self, submission: models.Submission):  # -> None:
        """Mark a ``Submission`` for sending."""
        submission.published = now()
        submission.status = SubmissionStatusEnum.PENDING
        submission.save()

    def submit_message(self, message: models.Message):  # -> models.Submission:
        """Creates a ``Submission`` instance for a given ``Message`` instance."""
        submission, _ = models.Submission.objects.get_or_create(message=message)
        return submission



from datetime import timedelta
from unittest.mock import Mock, patch, call

import pytest
from django.conf import settings
from django.test import override_settings
from django.urls import reverse
from django.utils.timezone import now

from mailinglist import models, services
from mailinglist.enum import SubmissionStatusEnum, SubscriptionStatusEnum


class TestTemplateSet:
    @patch("mailinglist.services.select_template")
    def test_get_template_no_mailing_list(self, p_select_template):
        ts = services.TemplateSet(mailing_list=None)._get_template()
        p_select_template.assert_called_once_with(
            [
                "mailinglist/email/global-deny/message.txt",
                "mailinglist/email/message.txt",
            ]
        )

    @patch("mailinglist.services.select_template")
    def test_get_template(self, p_select_template, mailing_list):
        services.TemplateSet(mailing_list=mailing_list)._get_template()
        p_select_template.assert_called_once_with(
            [
                f"mailinglist/email/{mailing_list.slug}/message.txt",
                "mailinglist/email/message.txt",
            ]
        )

    @patch("mailinglist.services.select_template")
    def test_get_template_suffix(self, p_select_template, mailing_list):
        services.TemplateSet(mailing_list=mailing_list)._get_template(suffix="html")
        p_select_template.assert_called_once_with(
            [
                f"mailinglist/email/{mailing_list.slug}/message.html",
                "mailinglist/email/message.html",
            ]
        )

    @patch("mailinglist.services.select_template")
    def test_get_template_action(self, p_select_template, mailing_list):
        services.TemplateSet(mailing_list=mailing_list)._get_template(_for="subscribe")
        p_select_template.assert_called_once_with(
            [
                f"mailinglist/email/{mailing_list.slug}/subscribe.txt",
                "mailinglist/email/subscribe.txt",
            ]
        )

    @patch(
        "mailinglist.services.TemplateSet._get_default_context",
        return_value={"omg": "yes"},
    )
    def test_render_to_dict(self, mailing_list):
        ts = services.TemplateSet(mailing_list=mailing_list)
        mock_template = Mock()
        mock_template.render.return_value = " ABC123 "
        ts._templates = {"somekey": mock_template}
        ret = ts.render_to_dict({"ctx_key": "render_context"})
        mock_template.render.assert_called_once_with(
            {"omg": "yes", "ctx_key": "render_context"}
        )
        assert ret["somekey"] == "ABC123"

    @patch(
        "mailinglist.services.TemplateSet._get_default_context",
        return_value={"omg": "yes"},
    )
    def test_render_to_dict_key_collision(self, mailing_list):
        ts = services.TemplateSet(mailing_list=mailing_list)
        mock_template = Mock()
        mock_template.render.return_value = " ABC123 "
        ts._templates = {"somekey": mock_template}
        ret = ts.render_to_dict({"omg": "render_context"})
        mock_template.render.assert_called_once_with({"omg": "render_context"})
        assert ret["somekey"] == "ABC123"

    @patch.object(services.TemplateSet, "_get_template")
    def test_templates(self, p_get_template, mailing_list):
        p_get_template.return_value = "generic return"
        ts = services.TemplateSet(mailing_list=mailing_list)
        ret = ts.templates
        assert ret == {
            "subject": "generic return",
            "body": "generic return",
            "html_body": "generic return",
        }
        p_get_template.assert_has_calls(
            [
                call(_for="message_subject"),
                call(_for="message"),
                call(_for="message", suffix="html"),
            ]
        )

    @patch.object(services.TemplateSet, "_get_template")
    def test_templates_no_html(self, p_get_template, mailing_list):
        mailing_list.send_html = False
        mailing_list.save()
        p_get_template.return_value = "generic return"
        ts = services.TemplateSet(mailing_list=mailing_list)
        ret = ts.templates
        assert ret == {
            "subject": "generic return",
            "body": "generic return",
        }
        p_get_template.assert_has_calls(
            [
                call(_for="message_subject"),
                call(_for="message"),
            ]
        )

    @patch.object(services.TemplateSet, "_get_template")
    def test_templates_caches(self, p_get_template):
        ts = services.TemplateSet()
        ts._templates = {"bypass": True}
        ret = ts.templates
        assert ret == {"bypass": True}
        p_get_template.assert_not_called()

    @override_settings(
        MAILINGLIST_DEFAULT_SENDER_NAME="Good Name",
        MAILINGLIST_BASE_URL="https://www.trustworthymailer.lol",
    )
    def test_get_default_context(self, mailing_list):
        ctx = services.TemplateSet(mailing_list=mailing_list)._get_default_context()
        assert ctx["mailing_list"] is mailing_list
        assert ctx["BASE_URL"] == "https://www.trustworthymailer.lol"
        assert ctx["DEFAULT_SENDER_NAME"] == "Good Name"

    @override_settings(
        MAILINGLIST_DEFAULT_SENDER_NAME="Good Name",
        MAILINGLIST_BASE_URL="https://www.trustworthymailer.lol",
    )
    def test_get_default_context_no_mailing_list(self):
        ctx = services.TemplateSet()._get_default_context()
        assert ctx["mailing_list"] is None
        assert ctx["BASE_URL"] == "https://www.trustworthymailer.lol"
        assert ctx["DEFAULT_SENDER_NAME"] == "Good Name"


class TestMessageService:
    @patch.object(
        services.MessageService, "_headers", return_value={"important-header": True}
    )
    @patch.object(services.TemplateSet, "render_to_dict")
    def test_prepare_kwargs(self, p_render, subscription, message):
        # ts = services.TemplateSet(mailing_list=message.mailing_list)
        p_render.return_value = {"stuff": True}
        ret = services.MessageService()._prepare_kwargs(
            subscription=subscription,
            message=message,
            template_set=services.TemplateSet(),
        )
        assert ret == {
            "stuff": True,
            "to": [subscription.user.email],
            "headers": {"important-header": True},
        }
        p_render.assert_called_once_with(
            context={"subscription": subscription, "message": message}
        )

    @patch.object(
        services.MessageService, "_headers", return_value={"important-header": True}
    )
    @patch.object(services.TemplateSet, "render_to_dict")
    def test_prepare_kwargs_no_message(self, p_render, subscription):
        # ts = services.TemplateSet(mailing_list=message.mailing_list)
        p_render.return_value = {"stuff": True}
        ret = services.MessageService()._prepare_kwargs(
            subscription=subscription,
            template_set=services.TemplateSet(),
        )
        assert ret == {
            "stuff": True,
            "to": [subscription.user.email],
            "headers": {"important-header": True},
        }
        p_render.assert_called_once_with(
            context={"subscription": subscription, "message": None}
        )

    def test_headers(self, subscription):
        ret = services.MessageService()._headers(subscription=subscription)
        for key in (
            "List-Help",
            "List-Unsubscribe",
            "List-Subscribe",
            "List-Post",
            "List-Owner",
            "List-Archive",
        ):
            assert key in ret, f"missing key {key}"
        for key in (
            "List-Help",
            "List-Unsubscribe",
            "List-Subscribe",
        ):
            assert subscription.token in ret[key]
        assert subscription.mailing_list.slug in ret["List-Archive"]

    def test_headers_no_mailing_list(self, subscription):
        subscription.mailing_list = None
        subscription.save()
        ret = services.MessageService()._headers(subscription=subscription)
        for key in (
            "List-Help",
            "List-Unsubscribe",
            "List-Subscribe",
            "List-Post",
            "List-Owner",
            "List-Archive",
        ):
            assert key in ret, f"missing key {key}"
        for key in (
            "List-Help",
            "List-Unsubscribe",
            "List-Subscribe",
        ):
            assert subscription.token in ret[key]
        assert ret["List-Archive"].endswith("mailinglist/archive/>")
        assert "," not in ret["List-Owner"]

    @patch.object(services.MessageService, "_prepare_kwargs")
    def test_prepare_message_kwargs(self, p_prepare):
        p_prepare.return_value = "generic return"
        ret = services.MessageService().prepare_message_kwargs(
            subscription=True, message=False, template_set=None
        )
        p_prepare.assert_called_once_with(
            subscription=True, message=False, template_set=None
        )
        assert ret == "generic return"

    @patch.object(services.MessageService, "_prepare_kwargs")
    def test_prepare_confirmation_kwargs(self, p_prepare):
        p_prepare.return_value = "generic return"
        ret = services.MessageService().prepare_confirmation_kwargs(
            subscription=True, template_set=None
        )
        p_prepare.assert_called_once_with(subscription=True, template_set=None)
        assert ret == "generic return"


class TestSubmissionService:
    def test_get_included_subscribers(self, active_subscription, submission):
        subscriptions = services.SubmissionService()._get_included_subscribers(
            submission
        )
        assert active_subscription in subscriptions

    def test_get_included_subscribers_excluded(self, active_subscription, submission):
        submission.exclude.add(active_subscription)
        subscriptions = services.SubmissionService()._get_included_subscribers(
            submission
        )
        assert active_subscription not in subscriptions

    def test_get_included_subscribers_global_deny(
        self, active_denied_subscription, submission
    ):
        subscriptions = services.SubmissionService()._get_included_subscribers(
            submission
        )
        assert active_denied_subscription not in subscriptions

    @override_settings(
        MAILINGLIST_EMAIL_DELAY=2,
        MAILINGLIST_BATCH_SIZE=200,
        MAILINGLIST_BATCH_DELAY=3,
    )
    @patch("time.sleep")
    def test_rate_limit(self, p_sleep):
        services.SubmissionService()._rate_limit(99)
        p_sleep.assert_called_once_with(2)

    @override_settings(
        MAILINGLIST_EMAIL_DELAY=2,
        MAILINGLIST_BATCH_SIZE=200,
        MAILINGLIST_BATCH_DELAY=3,
    )
    @patch("time.sleep")
    def test_rate_limit_batch(self, p_sleep):
        services.SubmissionService()._rate_limit(400)
        p_sleep.assert_called_once_with(3)

    @override_settings(
        MAILINGLIST_EMAIL_DELAY=2,
        MAILINGLIST_BATCH_SIZE=200,
        MAILINGLIST_BATCH_DELAY=None,
    )
    @patch("time.sleep")
    def test_rate_limit_batch_no_delay(self, p_sleep):
        services.SubmissionService()._rate_limit(400)
        p_sleep.assert_called_once_with(2)

    @override_settings(
        MAILINGLIST_EMAIL_DELAY=None,
        MAILINGLIST_BATCH_SIZE=200,
        MAILINGLIST_BATCH_DELAY=None,
    )
    @patch("time.sleep")
    def test_rate_limit_no_delay(self, p_sleep):
        services.SubmissionService()._rate_limit(400)
        p_sleep.assert_not_called()

    @patch.object(services.SubmissionService, "_send_message")
    def test_ensure_sent(self, p_send_message, active_subscription, submission):
        assert not models.Sending.objects.filter(
            submission=submission, subscription=active_subscription
        ).exists()
        did_send = services.SubmissionService()._ensure_sent(
            subscription=active_subscription, submission=submission
        )
        p_send_message.assert_called_once_with(
            message=submission.message, subscription=active_subscription
        )
        assert models.Sending.objects.filter(
            submission=submission, subscription=active_subscription
        ).exists()
        assert did_send
        models.Sending.objects.filter(
            submission=submission, subscription=active_subscription
        ).delete()

    @patch.object(services.SubmissionService, "_send_message")
    def test_ensure_sent_already_sent(
        self, p_send_message, active_subscription, submission
    ):
        models.Sending.objects.create(
            submission=submission, subscription=active_subscription
        )
        did_send = services.SubmissionService()._ensure_sent(
            subscription=active_subscription, submission=submission
        )
        p_send_message.assert_not_called()
        assert not did_send
        models.Sending.objects.filter(
            submission=submission, subscription=active_subscription
        ).delete()

    @patch.object(services.SubmissionService, "_send_message")
    def test_ensure_sent_send_failure(
        self, p_send_message, active_subscription, submission
    ):
        p_send_message.side_effect = Exception("boom")
        with pytest.raises(Exception):
            services.SubmissionService()._ensure_sent(
                subscription=active_subscription, submission=submission
            )
        assert not models.Sending.objects.filter(
            submission=submission, subscription=active_subscription
        ).exists()

    def test_get_outstanding_submissions_sending(self, submission):
        submission.status = SubmissionStatusEnum.PENDING
        submission.status = SubmissionStatusEnum.SENDING
        submission.save()
        outstanding = services.SubmissionService()._get_outstanding_submissions()
        assert submission in outstanding

    def test_get_outstanding_submissions_pending(self, submission):
        submission.status = SubmissionStatusEnum.PENDING
        submission.save()
        outstanding = services.SubmissionService()._get_outstanding_submissions()
        assert submission not in outstanding

    def test_get_outstanding_submissions_pending_future(self, submission):
        submission.status = SubmissionStatusEnum.PENDING
        submission.published = now() + timedelta(hours=1)
        submission.save()
        outstanding = services.SubmissionService()._get_outstanding_submissions()
        assert submission not in outstanding

    def test_get_outstanding_submissions_pending_ready(self, submission):
        submission.status = SubmissionStatusEnum.PENDING
        submission.published = now() - timedelta(hours=1)
        submission.save()
        outstanding = services.SubmissionService()._get_outstanding_submissions()
        assert submission in outstanding

    def test_get_outstanding_submissions_not_submitted(self, submission):
        outstanding = services.SubmissionService()._get_outstanding_submissions()
        assert submission not in outstanding

    @patch.object(services.SubmissionService, "_rate_limit")
    @patch.object(services.SubmissionService, "_ensure_sent")
    def test_process_submission_attachments(
        self,
        p_ensure_sent,
        p_rate_limit,
        submission,
        active_subscription,
        message_attachment,
    ):
        submission.status = SubmissionStatusEnum.PENDING
        submission.save()
        p_ensure_sent.return_value = True
        _send_count = services.SubmissionService().process_submission(submission)
        p_ensure_sent.assert_called_once()
        assert p_ensure_sent.call_args.kwargs["attachments"] == [message_attachment]
        p_rate_limit.assert_called_once_with(1)
        submission.refresh_from_db()
        assert submission.status == SubmissionStatusEnum.SENT
        assert _send_count == 1

    @patch.object(services.SubmissionService, "_rate_limit")
    @patch.object(services.SubmissionService, "_ensure_sent")
    def test_process_submission_template_set(
        self,
        p_ensure_sent,
        p_rate_limit,
        submission,
        active_subscription,
        message_attachment,
    ):
        submission.status = SubmissionStatusEnum.PENDING
        submission.save()
        p_ensure_sent.return_value = True
        _send_count = services.SubmissionService().process_submission(submission)
        p_ensure_sent.assert_called_once()
        assert (
            p_ensure_sent.call_args.kwargs["template_set"].mailing_list
            == submission.message.mailing_list
        )
        p_rate_limit.assert_called_once_with(1)
        submission.refresh_from_db()
        assert submission.status == SubmissionStatusEnum.SENT
        assert _send_count == 1

    @patch.object(services.SubmissionService, "_rate_limit")
    @patch.object(services.SubmissionService, "_ensure_sent")
    def test_process_submission(
        self, p_ensure_sent, p_rate_limit, submission, active_subscription
    ):
        submission.status = SubmissionStatusEnum.PENDING
        submission.save()
        p_ensure_sent.return_value = True
        _send_count = services.SubmissionService().process_submission(submission)
        p_ensure_sent.assert_called_once()
        p_rate_limit.assert_called_once_with(1)
        submission.refresh_from_db()
        assert submission.status == SubmissionStatusEnum.SENT
        assert _send_count == 1

    @patch.object(services.SubmissionService, "_rate_limit")
    @patch.object(services.SubmissionService, "_ensure_sent")
    def test_process_submission_send_count(
        self, p_ensure_sent, p_rate_limit, submission, active_subscription
    ):
        submission.status = SubmissionStatusEnum.PENDING
        submission.save()
        p_ensure_sent.return_value = True
        _send_count = services.SubmissionService().process_submission(
            submission, send_count=2
        )
        p_ensure_sent.assert_called_once()
        p_rate_limit.assert_called_once_with(3)
        submission.refresh_from_db()
        assert submission.status == SubmissionStatusEnum.SENT
        assert _send_count == 3

    @patch.object(services.SubmissionService, "_rate_limit")
    @patch.object(services.SubmissionService, "_ensure_sent")
    def test_process_submission_not_send(
        self, p_ensure_sent, p_rate_limit, submission, active_subscription
    ):
        submission.status = SubmissionStatusEnum.PENDING
        submission.save()
        p_ensure_sent.return_value = False
        _send_count = services.SubmissionService().process_submission(submission)
        p_ensure_sent.assert_called_once()
        p_rate_limit.assert_not_called()
        submission.refresh_from_db()
        assert submission.status == SubmissionStatusEnum.SENT
        assert _send_count == 0

    @patch.object(services.SubmissionService, "_get_included_subscribers")
    def test_process_submission_interrupt(self, p_get_included_subscribers, submission):
        p_get_included_subscribers.side_effect = Exception
        submission.status = SubmissionStatusEnum.PENDING
        submission.save()
        with pytest.raises(Exception):
            services.SubmissionService().process_submission(submission)
        submission.refresh_from_db()
        assert submission.status == SubmissionStatusEnum.SENDING

    @patch.object(services.SubmissionService, "process_submission")
    @patch.object(services.SubmissionService, "_get_outstanding_submissions")
    def test_process_submissions(
        self, p_get_outstanding_submissions, p_process_submission
    ):
        p_get_outstanding_submissions.return_value = [1, 2, 3]
        p_process_submission.return_value = 0
        services.SubmissionService().process_submissions()
        p_process_submission.assert_has_calls(
            [call(1, send_count=0), call(2, send_count=0), call(3, send_count=0)]
        )

    def test_publish(self, submission):
        assert submission.published is None
        assert submission.status == SubmissionStatusEnum.NEW
        services.SubmissionService().publish(submission)
        submission.refresh_from_db()
        assert submission.published is not None
        assert submission.status == SubmissionStatusEnum.PENDING

    def test_submit_message(self, message):
        assert not hasattr(message, "submission")
        services.SubmissionService().submit_message(message)
        message.refresh_from_db()
        assert hasattr(message, "submission")
        assert message.submission.status == SubmissionStatusEnum.NEW
        message.submission.delete()

    def test_submit_message_already_submitted(self, message, submission):
        assert hasattr(message, "submission")
        _submission = services.SubmissionService().submit_message(message)
        message.refresh_from_db()
        assert hasattr(message, "submission")
        assert submission == _submission
        assert message.submission.status == SubmissionStatusEnum.NEW
        assert models.Submission.objects.filter(message=message).count() == 1

    @patch.object(services.MessageService, "prepare_message_kwargs")
    @patch.object(services.hookset, "send_message")
    def test_send_message(
        self, p_send_message, p_prepare_message_kwargs, message, subscription
    ):
        p_prepare_message_kwargs.return_value = {"stuff": "yeah"}
        services.SubmissionService()._send_message(
            message=message, subscription=subscription, template_set=None
        )
        p_prepare_message_kwargs.assert_called_once_with(
            message=message, subscription=subscription, template_set=None
        )
        p_send_message.assert_called_once_with(
            from_email=subscription.mailing_list.sender_tag,
            stuff="yeah",
        )

    @patch.object(services.MessageService, "prepare_message_kwargs")
    @patch.object(services.hookset, "send_message")
    def test_send_message_extra_kwargs(
        self, p_send_message, p_prepare_message_kwargs, message, subscription
    ):
        p_prepare_message_kwargs.return_value = {"stuff": "yeah"}
        services.SubmissionService()._send_message(
            message=message, subscription=subscription, template_set=None, more=True
        )
        p_prepare_message_kwargs.assert_called_once_with(
            message=message, subscription=subscription, template_set=None
        )
        p_send_message.assert_called_once_with(
            from_email=subscription.mailing_list.sender_tag, stuff="yeah", more=True
        )


class TestSubscriptionService:
    @patch("mailinglist.services.randint", Mock(return_value=3))
    def test_rotate_token(self, subscription):
        _old_token = subscription.token
        services.SubscriptionService()._rotate_token(subscription)
        subscription.refresh_from_db()
        _hash = str(
            hash(
                f"{subscription.user.email[0:3]}"
                f"{subscription.mailing_list.name}"
                f"{subscription.user.email[3:0]}"
            )
        )

        assert subscription.token.endswith(_hash)
        assert _old_token != subscription.token

    @patch("mailinglist.services.randint", Mock(return_value=3))
    def test_new_subscription(self, mailing_list, user):
        subscription = services.SubscriptionService()._new_subscription(
            user=user, mailing_list=mailing_list
        )
        _hash = str(hash(f"{user.email[0:3]}{mailing_list.name}{user.email[3:0]}"))
        assert subscription.token.endswith(_hash)
        assert subscription.status == SubscriptionStatusEnum.PENDING

    def test_update_subscription_status_no_change(self, active_subscription):
        assert active_subscription.status == SubscriptionStatusEnum.SUBSCRIBED
        qs = models.SubscriptionChange.objects.filter(subscription=active_subscription)
        assert qs.count() == 0
        services.SubscriptionService()._update_subscription_status(
            subscription=active_subscription,
            to_status=SubscriptionStatusEnum.SUBSCRIBED,
        )
        active_subscription.refresh_from_db()
        assert active_subscription.status == SubscriptionStatusEnum.SUBSCRIBED
        assert qs.count() == 0

    def test_update_subscription_status_subscribe(self, subscription):
        assert subscription.status != SubscriptionStatusEnum.SUBSCRIBED
        qs = models.SubscriptionChange.objects.filter(subscription=subscription)
        assert qs.count() == 0
        services.SubscriptionService()._update_subscription_status(
            subscription=subscription,
            to_status=SubscriptionStatusEnum.SUBSCRIBED,
        )
        subscription.refresh_from_db()
        assert subscription.status == SubscriptionStatusEnum.SUBSCRIBED
        assert qs.count() == 1
        assert qs[0].to_status == SubscriptionStatusEnum.SUBSCRIBED

    def test_update_subscription_status_unsubscribe(self, subscription):
        assert subscription.status != SubscriptionStatusEnum.UNSUBSCRIBED
        qs = models.SubscriptionChange.objects.filter(subscription=subscription)
        assert qs.count() == 0
        services.SubscriptionService()._update_subscription_status(
            subscription=subscription,
            to_status=SubscriptionStatusEnum.UNSUBSCRIBED,
        )
        subscription.refresh_from_db()
        assert subscription.status == SubscriptionStatusEnum.UNSUBSCRIBED
        assert qs.count() == 1
        assert qs[0].to_status == SubscriptionStatusEnum.UNSUBSCRIBED

    @patch.object(services.SubscriptionService, "_send_subscription_confirmation")
    def test_subscribe(self, p_confirm, user, mailing_list):
        subscription = services.SubscriptionService().subscribe(
            user=user, mailing_list=mailing_list
        )
        assert subscription.status == SubscriptionStatusEnum.PENDING
        p_confirm.assert_called_once_with(subscription)

    @patch.object(services.SubscriptionService, "_send_subscription_confirmation")
    def test_subscribe_force_confirm(self, p_confirm, user, mailing_list):
        subscription = services.SubscriptionService().subscribe(
            user=user, mailing_list=mailing_list, force_confirm=True
        )
        assert subscription.status == SubscriptionStatusEnum.SUBSCRIBED
        p_confirm.assert_not_called()

    @patch.object(services.SubscriptionService, "_new_subscription")
    def test_subscribe_exists(self, p_new_subscription, active_subscription):
        services.SubscriptionService().subscribe(
            user=active_subscription.user, mailing_list=active_subscription.mailing_list
        )
        active_subscription.refresh_from_db()
        p_new_subscription.assert_not_called()
        assert active_subscription.status == SubscriptionStatusEnum.SUBSCRIBED

    def test_subscribe_deny(self, denied_user, mailing_list):
        subscription = services.SubscriptionService().subscribe(
            user=denied_user, mailing_list=mailing_list
        )
        assert subscription is None
        assert not models.Subscription.objects.filter(user=denied_user).exists()

    @override_settings(MAILINGLIST_CONFIRM_EMAIL_SUBSCRIBE=False)
    def test_subscribe_noconfirm(self, user, mailing_list):
        subscription = services.SubscriptionService().subscribe(
            user=user, mailing_list=mailing_list
        )
        assert subscription.status == SubscriptionStatusEnum.SUBSCRIBED

    def test_unsubscribe(self, active_subscription):
        assert active_subscription.status != SubscriptionStatusEnum.UNSUBSCRIBED
        services.SubscriptionService().unsubscribe(token=active_subscription.token)
        active_subscription.refresh_from_db()
        assert active_subscription.status == SubscriptionStatusEnum.UNSUBSCRIBED

    def test_unsubscribe_bad_token(self, active_subscription):
        assert active_subscription.status != SubscriptionStatusEnum.UNSUBSCRIBED
        ret = services.SubscriptionService().unsubscribe(
            token=active_subscription.token + "a"
        )
        assert ret is None
        active_subscription.refresh_from_db()
        assert active_subscription.status == SubscriptionStatusEnum.SUBSCRIBED

    def test_confirm_subscription(self, subscription):
        assert subscription.status != SubscriptionStatusEnum.SUBSCRIBED
        services.SubscriptionService().confirm_subscription(token=subscription.token)
        subscription.refresh_from_db()
        assert subscription.status == SubscriptionStatusEnum.SUBSCRIBED

    @override_settings(
        MAILINGLIST_CONFIRM_EMAIL_SUBSCRIBE=False,
        MAILINGLIST_BASE_URL="https://www.trustworthymailer.lol",
    )
    @patch.object(services.hookset, "send_message")
    def test_new_subscription_no_mailing_list(self, p_send, user):
        assert not models.GlobalDeny.objects.filter(user=user).exists()
        subscription = services.SubscriptionService().subscribe(
            user=user, mailing_list=None
        )
        confirm_url = reverse(
            "mailinglist:subscribe_confirm", kwargs={"token": subscription.token}
        )
        full_url = f"https://www.trustworthymailer.lol{confirm_url}"
        # sends email with token always!
        p_send.assert_called_once()
        assert full_url in p_send.call_args.kwargs["body"]
        assert full_url in p_send.call_args.kwargs["html_body"]
        assert not models.GlobalDeny.objects.filter(user=user).exists()
        # token redeems by putting user in global deny
        services.SubscriptionService().confirm_subscription(token=subscription.token)
        assert models.GlobalDeny.objects.filter(user=user).exists()

    def test_force_subscribe(self, user, mailing_list):
        assert not user.subscriptions.all().exists()
        subscription = services.SubscriptionService().force_subscribe(
            user=user, mailing_list=mailing_list
        )
        assert user.subscriptions.all().exists()
        assert subscription.user == user
        assert subscription.mailing_list == mailing_list
        assert subscription.status == SubscriptionStatusEnum.SUBSCRIBED



import pytest
from unittest.mock import call, patch, Mock
from django.http import Http404
from mailinglist import admin
from django.contrib import admin as admin_site
from django.contrib.auth.models import Permission

from django.shortcuts import reverse
from mailinglist.models import Subscription, Submission
from mailinglist.enum import SubmissionStatusEnum
from mailinglist import services


@pytest.fixture
def inner_mock():
    return Mock()


@pytest.fixture
def subscription_admin(inner_mock):
    return admin.SubscriptionAdmin(
        Subscription, Mock(admin_view=Mock(return_value=inner_mock))
    )


@pytest.fixture
def submission_admin(inner_mock):
    return admin.SubmissionAdmin(
        Submission, Mock(admin_view=Mock(return_value=inner_mock))
    )


@pytest.fixture
def subscription_import_response(admin_client, mailing_list, address_file):
    return admin_client.post(
        reverse("admin:mailinglist_subscription_import"),
        {
            "mailing_list": mailing_list.pk,
            "address_file": address_file,
            "ignore_errors": True,
        },
        follow=True,
    )


@pytest.fixture
def confirmed_subscription_import_response(admin_client, subscription_import_response):
    return admin_client.post(
        reverse("admin:mailinglist_subscription_import_confirm"),
        {"confirm": True},
        follow=True,
    )


class TestExtendibleModelAdminMixin:
    # mixin is too abstract, have to test concrete class
    @patch.object(admin.SubscriptionAdmin, "get_queryset")
    def test_get_object_missing(self, p_get_queryset):
        p_get_queryset.side_effect = Subscription.DoesNotExist
        with pytest.raises(Http404):
            admin.SubscriptionAdmin(Subscription, admin_site)._getobj(None, None)

    @patch.object(admin.SubscriptionAdmin, "get_queryset")
    def test_get_object_present(self, p_get_queryset):
        p_get_queryset.return_value = Mock(get=Mock(return_value="something"))
        assert (
            admin.SubscriptionAdmin(Subscription, admin_site)._getobj(None, "25")
            == "something"
        )

    def test_view_name(self):
        assert (
            admin.SubscriptionAdmin(Subscription, admin_site)._view_name("snut")
            == "mailinglist_subscription_snut"
        )

    def test_wrap(self, subscription_admin, inner_mock):
        _callable = subscription_admin._wrap(subscription_admin.subscribers_import)
        _callable("wild")
        inner_mock.assert_called_once_with("wild")


class TestImmutableAdminMixin:
    def test_add_permission(self):
        assert not admin.ImmutableAdminMixin().has_add_permission(None, None)

    def test_change_permission(self):
        assert not admin.ImmutableAdminMixin().has_change_permission(None, None)

    def test_delete_permission(self):
        assert not admin.ImmutableAdminMixin().has_delete_permission(None, None)


class TestUnchangingAdminMixin:
    def test_change_permission(self):
        assert not admin.UnchangingAdminMixin().has_change_permission(None, None)


class TestSubscriptionAdmin:
    @patch.object(services.SubscriptionService, "subscribe")
    def test_save_object_no_change(self, p_subscribe, subscription_admin, subscription):
        subscription_admin.save_model(None, subscription, None, False)
        p_subscribe.assert_called_once_with(
            user=subscription.user, mailing_list=subscription.mailing_list
        )

    @patch.object(services.SubscriptionService, "subscribe")
    def test_save_object(self, p_subscribe, subscription_admin, subscription):
        subscription_admin.save_model(None, subscription, None, True)
        p_subscribe.assert_not_called()

    def test_has_change_permission(self, subscription_admin):
        assert not subscription_admin.has_change_permission(None)

    @patch.object(services.SubscriptionService, "_confirm_subscription")
    @patch.object(admin.SubscriptionAdmin, "message_user")
    def test_make_subscribed(self, p_message, p_confirm, subscription_admin):
        subscription_admin.make_subscribed(None, ["a", "b", "c"])
        p_confirm.assert_has_calls([call("a"), call("b"), call("c")])
        p_message.assert_called_once_with(
            None, "3 users have been successfully subscribed."
        )

    @patch.object(services.SubscriptionService, "_confirm_unsubscription")
    @patch.object(admin.SubscriptionAdmin, "message_user")
    def test_make_unsubscribed(self, p_message, p_confirm, subscription_admin):
        subscription_admin.make_unsubscribed(None, ["a", "b", "c"])
        p_confirm.assert_has_calls([call("a"), call("b"), call("c")])
        p_message.assert_called_once_with(
            None, "3 users have been successfully unsubscribed."
        )

    def test_subscribers_import(self, subscription_import_response):
        assert b"<h1>Confirm import</h1>" in subscription_import_response.content
        assert (
            b"<li>&lt;last@email.com&gt; Seymore Lastman</li>"
            in subscription_import_response.content
        )

    def test_subscribers_import_confirm(self, confirmed_subscription_import_response):
        assert (
            b"1 subscriptions have been added."
            in confirmed_subscription_import_response.content
        )

    def test_subscribers_import_bad_post(self, admin_client, mailing_list):
        response = admin_client.post(
            reverse("admin:mailinglist_subscription_import"),
            {
                "mailing_list": mailing_list.pk,
                "ignore_errors": True,
            },
            follow=True,
        )
        print(response.content)
        assert (
            b'<ul class="errorlist"><li>This field is required.</li></ul>'
            in response.content
        )

    def test_subscribers_import_denied(self, admin_client, admin_user):
        admin_user.is_superuser = False
        admin_user.save()
        response = admin_client.get(reverse("admin:mailinglist_subscription_import"))

        assert response.status_code == 403

    def test_subscribers_import_tight_permission(self, admin_client, admin_user):
        admin_user.is_superuser = False
        admin_user.user_permissions.add(
            Permission.objects.get(codename="add_subscription")
        )
        admin_user.save()
        response = admin_client.get(reverse("admin:mailinglist_subscription_import"))

        assert response.status_code == 200

    def test_subscribers_import_get(self, admin_client):
        response = admin_client.get(
            reverse("admin:mailinglist_subscription_import"), follow=True
        )
        assert b"<h1>Import addresses</h1>" in response.content

    def test_subscribers_import_confirm_skipped(self, admin_client):
        response = admin_client.get(
            reverse("admin:mailinglist_subscription_import_confirm")
        )
        assert response.status_code == 302
        assert response.url == reverse("admin:mailinglist_subscription_import")

    def test_subscribers_import_confirm_bad_post(
        self, admin_client, subscription_import_response
    ):
        response = admin_client.post(
            reverse("admin:mailinglist_subscription_import_confirm"),
            {"confirm": False},
            follow=False,
        )
        print(response.content)
        assert b"<li>You should confirm in order to continue.</li>" in response.content


class TestMessageAdmin:
    def test_preview(self, admin_client, message, message_part):
        response = admin_client.get(
            reverse(
                "admin:mailinglist_message_preview", kwargs={"object_id": message.pk}
            )
        )
        assert response.status_code == 200
        print(response.content)
        assert (
            str.encode(
                f"<iframe src ='/admin/mailinglist/message/{message.pk}/preview/html/'"
            )
            in response.content
        )
        assert (
            str.encode(
                f"<iframe src ='/admin/mailinglist/message/{message.pk}/preview/text/'"
            )
            in response.content
        )

    def test_preview_no_html(self, admin_client, message, message_part):
        message.mailing_list.send_html = False
        message.mailing_list.save()
        response = admin_client.get(
            reverse(
                "admin:mailinglist_message_preview", kwargs={"object_id": message.pk}
            )
        )
        assert response.status_code == 200
        print(response.content)
        assert (
            str.encode(
                f"<iframe src ='/admin/mailinglist/message/{message.pk}/preview/html/'"
            )
            not in response.content
        )
        assert (
            str.encode(
                f"<iframe src ='/admin/mailinglist/message/{message.pk}/preview/text/'"
            )
            in response.content
        )

    def test_preview_html(self, admin_client, message, message_part):
        response = admin_client.get(
            reverse(
                "admin:mailinglist_message_preview_html",
                kwargs={"object_id": message.pk},
            )
        )
        assert response.status_code == 200
        print(response.content)
        assert b"<strong>exquisite</strong>" in response.content
        assert b"<li>and lists</li>" in response.content

    def test_preview_text(self, admin_client, message, message_part):
        response = admin_client.get(
            reverse(
                "admin:mailinglist_message_preview_text",
                kwargs={"object_id": message.pk},
            )
        )
        assert response.status_code == 200
        print(response.content)
        assert b"**exquisite**" in response.content
        assert b"* and lists" in response.content

    def test_submit(self, admin_client, message, message_part):
        assert not hasattr(message, "submission")
        response = admin_client.get(
            reverse(
                "admin:mailinglist_message_submit",
                kwargs={"object_id": message.pk},
            ),
            follow=False,
        )
        message.refresh_from_db()
        assert hasattr(message, "submission")
        assert response.status_code == 302
        assert response.url == reverse(
            "admin:mailinglist_submission_change", args=[message.submission.pk]
        )
        Submission.objects.filter(pk=message.submission.pk).delete()


class TestSubmissionAdmin:
    def test_submit(self, admin_client, submission):
        assert submission.status == SubmissionStatusEnum.NEW
        assert submission.published is None
        response = admin_client.get(
            reverse(
                "admin:mailinglist_submission_publish",
                kwargs={"object_id": submission.pk},
            ),
            follow=False,
        )
        submission.refresh_from_db()
        assert submission.status == SubmissionStatusEnum.PENDING
        assert submission.published is not None
        assert response.status_code == 302
        assert response.url == reverse("admin:mailinglist_submission_changelist")

    @patch.object(services.SubmissionService, "publish")
    @patch.object(admin.SubmissionAdmin, "message_user")
    def test_publish_action(self, p_message, p_publish, submission_admin):
        submission_admin.publish(None, ["a", "b", "c"])
        p_publish.assert_has_calls([call("a"), call("b"), call("c")])
        p_message.assert_called_once_with(None, "3 submissions have been published.")



from mailinglist import enum, forms, models


class TestSubscriptionForm:
    def test_initial_not_subscribed(self, subscription):
        form = forms.SubscriptionForm(user=subscription.user)
        _list_key = f"mailing-list_{subscription.mailing_list.pk}"
        assert _list_key in form.fields
        assert not form.fields[_list_key].initial

    def test_initial_subscribed(self, active_subscription):
        form = forms.SubscriptionForm(user=active_subscription.user)
        _list_key = f"mailing-list_{active_subscription.mailing_list.pk}"
        assert _list_key in form.fields
        assert form.fields[_list_key].initial

    def test_initial_global_deny(self, denied_subscription):
        form = forms.SubscriptionForm(user=denied_subscription.user)
        _list_key = f"mailing-list_{denied_subscription.mailing_list.pk}"
        assert _list_key in form.fields
        assert not form.fields[_list_key].initial
        assert form.fields["global-deny"].initial

    def test_can_subscribe(self, subscription):
        assert subscription.status == enum.SubscriptionStatusEnum.PENDING
        _list_key = f"mailing-list_{subscription.mailing_list.pk}"
        form = forms.SubscriptionForm(user=subscription.user, data={_list_key: True})
        form.is_valid()
        form.save()
        subscription.refresh_from_db()
        assert subscription.status == enum.SubscriptionStatusEnum.SUBSCRIBED

    def test_can_unsubscribe(self, active_subscription):
        assert active_subscription.status == enum.SubscriptionStatusEnum.SUBSCRIBED
        _list_key = f"mailing-list_{active_subscription.mailing_list.pk}"
        form = forms.SubscriptionForm(
            user=active_subscription.user, data={_list_key: False}
        )
        form.is_valid()
        form.save()
        active_subscription.refresh_from_db()
        assert active_subscription.status == enum.SubscriptionStatusEnum.UNSUBSCRIBED

    def test_nothing_changes(self, active_subscription):
        assert active_subscription.status == enum.SubscriptionStatusEnum.SUBSCRIBED
        _list_key = f"mailing-list_{active_subscription.mailing_list.pk}"
        form = forms.SubscriptionForm(
            user=active_subscription.user, data={_list_key: True}
        )
        form.is_valid()
        form.save()
        active_subscription.refresh_from_db()
        assert active_subscription.status == enum.SubscriptionStatusEnum.SUBSCRIBED

    def test_can_global_deny(self, active_subscription, user):
        assert active_subscription.status == enum.SubscriptionStatusEnum.SUBSCRIBED
        _list_key = f"mailing-list_{active_subscription.mailing_list.pk}"
        form = forms.SubscriptionForm(
            user=active_subscription.user, data={_list_key: True, "global-deny": True}
        )
        form.is_valid()
        form.save()
        assert models.GlobalDeny.objects.filter(user=user).exists()

    def test_can_un_global_deny(self, active_denied_subscription, denied_user):
        assert (
            active_denied_subscription.status == enum.SubscriptionStatusEnum.SUBSCRIBED
        )
        assert models.GlobalDeny.objects.filter(user=denied_user).exists()
        _list_key = f"mailing-list_{active_denied_subscription.mailing_list.pk}"
        form = forms.SubscriptionForm(
            user=active_denied_subscription.user,
            data={_list_key: True, "global-deny": False},
        )
        form.is_valid()
        form.save()
        assert not models.GlobalDeny.objects.filter(user=denied_user).exists()

    def test_can_subscribe_nothing_happens(self, subscription):
        assert subscription.status == enum.SubscriptionStatusEnum.PENDING
        _list_key = f"mailing-list_{subscription.mailing_list.pk}"
        form = forms.SubscriptionForm(user=subscription.user, data={_list_key: False})
        form.is_valid()
        form.save(commit=False)
        subscription.refresh_from_db()
        assert subscription.status == enum.SubscriptionStatusEnum.PENDING

    def test_no_mailing_lists(self, user):
        form = forms.SubscriptionForm(user=user)
        assert len(form.fields) == 1
        assert "global-deny" in form.fields

    def test_no_mailing_lists_can_global_deny(self, user):
        assert not models.GlobalDeny.objects.filter(user=user).exists()
        form = forms.SubscriptionForm(user=user, data={"global-deny": True})
        assert len(form.fields) == 1
        assert "global-deny" in form.fields
        form.is_valid()
        form.save()
        assert models.GlobalDeny.objects.filter(user=user).exists()



import logging

logger = logging.getLogger(__name__)
from functools import update_wrapper

from django.contrib import admin, messages
from django.contrib.admin.utils import unquote
from django.core.exceptions import PermissionDenied
from django.http import Http404, HttpResponse, HttpResponseRedirect
from django.shortcuts import render
from django.urls import path, reverse
from django.utils.encoding import force_str
from django.views.decorators.clickjacking import xframe_options_sameorigin

from mailinglist import models
from mailinglist.admin_forms import ConfirmForm, ImportForm, SubmissionModelForm
from mailinglist.services import MessageService, SubmissionService, SubscriptionService


class ExtendibleModelAdminMixin:
    def _getobj(self, request, object_id):
        opts = self.model._meta

        try:
            obj = self.get_queryset(request).get(pk=unquote(object_id))
        except self.model.DoesNotExist:
            # Don't raise Http404 just yet, because we haven't checked
            # permissions yet. We don't want an unauthenticated user to
            # be able to determine whether a given object exists.
            obj = None

        if obj is None:
            raise Http404(
                f"{force_str(opts.verbose_name)} object with primary key "
                f"'{force_str(object_id)}' does not exist."
            )

        return obj

    def _wrap(self, view):
        def wrapper(*args, **kwargs):
            return self.admin_site.admin_view(view)(*args, **kwargs)

        return update_wrapper(wrapper, view)

    def _view_name(self, name):
        return f"{self.model._meta.app_label}_{self.model._meta.model_name}_{name}"


class ImmutableAdminMixin:
    can_delete = False

    def has_add_permission(self, request, obj):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


class ImmutableTabluarInline(ImmutableAdminMixin, admin.TabularInline):
    pass


@admin.register(models.MailingList)
class MailingListAdmin(admin.ModelAdmin):
    model = models.MailingList
    list_display = ("name", "slug", "visible")
    prepopulated_fields = {"slug": ("name",)}


class SubscriptionChangeInline(ImmutableTabluarInline):
    model = models.SubscriptionChange
    readonly_fields = ["from_status", "to_status", "changed"]


class SendingInline(ImmutableTabluarInline):
    model = models.Sending
    readonly_fields = ["submission", "sent"]


@admin.register(models.Subscription)
class SubscriptionAdmin(ExtendibleModelAdminMixin, admin.ModelAdmin):
    model = models.Subscription
    readonly_fields = ("token", "status")
    list_display = ("pk", "user", "mailing_list", "status")
    list_filter = ("mailing_list", "status")
    inlines = (SubscriptionChangeInline, SendingInline)
    actions = ("make_subscribed", "make_unsubscribed")

    def save_model(self, request, obj, form, change):
        if not change:
            SubscriptionService().subscribe(
                user=obj.user, mailing_list=obj.mailing_list
            )

    def has_change_permission(self, request, obj=None):
        return False

    def make_subscribed(self, request, queryset):
        # rows_updated = queryset.update(subscribed=True)
        service = SubscriptionService()
        rows_updated = 0
        for row in queryset:
            service._confirm_subscription(row)
            rows_updated += 1
        self.message_user(
            request, f"{rows_updated} users have been successfully subscribed."
        )

    make_subscribed.short_description = "Subscribe selected users"

    def make_unsubscribed(self, request, queryset):
        service = SubscriptionService()
        rows_updated = 0
        for row in queryset:
            service._confirm_unsubscription(row)
            rows_updated += 1
        self.message_user(
            request, f"{rows_updated} users have been successfully unsubscribed."
        )

    make_unsubscribed.short_description = "Unsubscribe selected users"

    """ Views """

    def subscribers_import(self, request):
        if not request.user.has_perm("mailinglist.add_subscription"):
            raise PermissionDenied()
        if request.POST:
            form = ImportForm(request.POST, request.FILES)
            if form.is_valid():
                request.session["addresses"] = form.get_addresses()
                request.session["mailing_list_pk"] = form.cleaned_data[
                    "mailing_list"
                ].pk

                confirm_url = reverse("admin:mailinglist_subscription_import_confirm")
                return HttpResponseRedirect(confirm_url)
        else:
            form = ImportForm()

        return render(
            request,
            "admin/mailinglist/subscription/import_form.html",
            {"form": form},
        )

    def subscribers_import_confirm(self, request):
        # If no addresses are in the session, start all over.

        if "addresses" not in request.session:
            import_url = reverse("admin:mailinglist_subscription_import")
            return HttpResponseRedirect(import_url)

        addresses = request.session["addresses"]
        mailing_list = models.MailingList.objects.get(
            pk=request.session["mailing_list_pk"]
        )

        logger.debug("Confirming addresses: %s", addresses)

        if request.POST:
            form = ConfirmForm(request.POST)
            if form.is_valid():
                try:
                    service = SubscriptionService()
                    for _user in addresses.values():
                        user = service.create_user(**_user)
                        service.force_subscribe(user=user, mailing_list=mailing_list)
                finally:
                    del request.session["addresses"]
                    del request.session["mailing_list_pk"]

                messages.success(
                    request,
                    f"{len(addresses)} subscriptions have been added.",
                )

                changelist_url = reverse("admin:mailinglist_subscription_changelist")
                return HttpResponseRedirect(changelist_url)
        else:
            form = ConfirmForm()

        return render(
            request,
            "admin/mailinglist/subscription/confirm_import_form.html",
            {"form": form, "subscribers": addresses},
        )

    """ URLs """

    def get_urls(self):
        urls = super().get_urls()

        my_urls = [
            path(
                "import/",
                self._wrap(self.subscribers_import),
                name=self._view_name("import"),
            ),
            path(
                "import/confirm/",
                self._wrap(self.subscribers_import_confirm),
                name=self._view_name("import_confirm"),
            ),
        ]

        return my_urls + urls


class UnchangingAdminMixin:
    def has_change_permission(self, request, obj=None):
        return False


@admin.register(models.GlobalDeny)
class GlobalDenyAdmin(UnchangingAdminMixin, admin.ModelAdmin):
    model = models.GlobalDeny
    list_display = ("pk", "user", "created")
    date_hierarchy = "created"


class MessagePartInline(admin.StackedInline):
    model = models.MessagePart
    extra = 2


class MessageAttachmentInline(UnchangingAdminMixin, admin.TabularInline):
    model = models.MessageAttachment
    extra = 1
    fields = ("file",)


@admin.register(models.Message)
class MessageAdmin(ExtendibleModelAdminMixin, admin.ModelAdmin):
    model = models.Message
    prepopulated_fields = {"slug": ("title",)}
    list_display = ("title", "slug", "mailing_list", "created")
    inlines = (MessagePartInline, MessageAttachmentInline)

    """ Views """

    def preview(self, request, object_id):
        return render(
            request,
            "admin/mailinglist/preview.html",
            {
                "message": self._getobj(request, object_id),
                "attachments": models.MessageAttachment.objects.filter(
                    message_id=object_id
                ),
            },
        )

    @xframe_options_sameorigin
    def preview_html(self, request, object_id):
        message = self._getobj(request, object_id)
        html_body = MessageService().prepare_message_preview_html(message=message)
        return HttpResponse(html_body)

    @xframe_options_sameorigin
    def preview_text(self, request, object_id):
        message = self._getobj(request, object_id)
        body = MessageService().prepare_message_preview(message=message)
        return HttpResponse(body, content_type="text/plain")

    def submit(self, request, object_id):
        submission = SubmissionService().submit_message(
            self._getobj(request, object_id)
        )

        change_url = reverse(
            "admin:mailinglist_submission_change", args=[submission.id]
        )

        return HttpResponseRedirect(change_url)

    """ URLs """

    def get_urls(self):
        urls = super().get_urls()

        my_urls = [
            path(
                "<object_id>/preview/",
                self._wrap(self.preview),
                name=self._view_name("preview"),
            ),
            path(
                "<object_id>/preview/html/",
                self._wrap(self.preview_html),
                name=self._view_name("preview_html"),
            ),
            path(
                "<object_id>/preview/text/",
                self._wrap(self.preview_text),
                name=self._view_name("preview_text"),
            ),
            path(
                "<object_id>/submit/",
                self._wrap(self.submit),
                name=self._view_name("submit"),
            ),
        ]

        return my_urls + urls


@admin.register(models.Submission)
class SubmissionAdmin(ExtendibleModelAdminMixin, admin.ModelAdmin):
    form = SubmissionModelForm
    model = models.Submission
    list_display = ("__str__", "status", "published")
    readonly_fields = ("published", "status")
    exclude = ("sendings",)
    actions = ("publish",)

    def publish(self, request, queryset):
        service = SubmissionService()
        rows_updated = 0
        for row in queryset:
            service.publish(row)
            rows_updated += 1
        self.message_user(request, f"{rows_updated} submissions have been published.")

    def publish_view(self, request, object_id):
        service = SubmissionService()
        service.publish(self._getobj(request, object_id))

        messages.success(request, f"Submission {object_id} has been published.")

        changelist_url = reverse("admin:mailinglist_submission_changelist")
        return HttpResponseRedirect(changelist_url)

    def get_urls(self):
        urls = super().get_urls()

        my_urls = [
            path(
                "<object_id>/publish/",
                self._wrap(self.publish_view),
                name=self._view_name("publish"),
            ),
        ]

        return my_urls + urls



# Generated by Django 4.2.7 on 2023-11-23 17:36

from django.db import migrations, models
import django.db.models.deletion
import mailinglist.models


class Migration(migrations.Migration):
    dependencies = [
        ("mailinglist", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="MessageAttachment",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "file",
                    models.FileField(
                        upload_to=mailinglist.models.attachment_upload_to,
                        validators=[mailinglist.models.hookset_validation_wrapper],
                        verbose_name="attachment",
                    ),
                ),
                ("filename", models.CharField(max_length=256)),
                (
                    "message",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="attachments",
                        to="mailinglist.message",
                    ),
                ),
            ],
        ),
    ]



import warnings

from setuptools import find_packages, setup

try:
    README = open("README.md").read() + "\n\n"
    README += open("CHANGELOG.md").read()
except:  # noqa: E722
    warnings.warn("Could not read README.md and/or CHANGELOG.md")
    README = None

version = __import__("mailinglist").__version__

setup(
    name="django-mailinglist",
    version=version,
    description=(
        "Django app for managing multiple mailing lists with both "
        "plaintext as well as HTML templates (facilitated by Markdown)."
    ),
    long_description=README,
    long_description_content_type="text/markdown",
    install_requires=[
        "Django>=3.2.0",
        "Markdown>=3.3.0",
        "django-enumfield>=3.0",
        "django-appconf>=1.0.0",
    ],
    author="Paul Stiverson",
    author_email="paul@thismatters.net",
    url="http://github.com/thismatters/django-mailinglist/",
    packages=find_packages(exclude=("tests", "test_project")),
    include_package_data=True,
    python_requires=">=3.8",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Framework :: Django :: 3.2",
        "Framework :: Django :: 4.1",
        "Framework :: Django :: 4.2",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Utilities",
    ],
)



# Generated by Django 4.2.7 on 2023-11-19 13:54

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django_enumfield.db.fields
import mailinglist.enum


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.MAILINGLIST_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="MailingList",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=128)),
                ("slug", models.SlugField(unique=True)),
                ("email", models.EmailField(help_text="Sender e-mail", max_length=254)),
                ("sender", models.CharField(help_text="Sender name", max_length=200)),
                ("visible", models.BooleanField(db_index=True, default=True)),
                (
                    "send_html",
                    models.BooleanField(
                        default=True,
                        help_text="Whether or not to send HTML versions of e-mails.",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Message",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("title", models.CharField(max_length=128)),
                ("slug", models.SlugField()),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("modified", models.DateTimeField(auto_now=True)),
                (
                    "mailing_list",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="messages",
                        to="mailinglist.mailinglist",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Sending",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("sent", models.DateTimeField(auto_now_add=True)),
            ],
        ),
        migrations.CreateModel(
            name="Subscription",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("token", models.CharField(max_length=45)),
                (
                    "status",
                    django_enumfield.db.fields.EnumField(
                        default=0, enum=mailinglist.enum.SubscriptionStatusEnum
                    ),
                ),
                (
                    "mailing_list",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="subscriptions",
                        to="mailinglist.mailinglist",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="subscriptions",
                        to=settings.MAILINGLIST_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="SubscriptionChange",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "from_status",
                    django_enumfield.db.fields.EnumField(
                        blank=True,
                        default=0,
                        enum=mailinglist.enum.SubscriptionStatusEnum,
                        null=True,
                    ),
                ),
                (
                    "to_status",
                    django_enumfield.db.fields.EnumField(
                        default=0, enum=mailinglist.enum.SubscriptionStatusEnum
                    ),
                ),
                ("changed", models.DateTimeField(auto_now_add=True)),
                (
                    "subscription",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="mailinglist.subscription",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Submission",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("published", models.DateTimeField(blank=True, null=True)),
                (
                    "status",
                    django_enumfield.db.fields.EnumField(
                        default=0, enum=mailinglist.enum.SubmissionStatusEnum
                    ),
                ),
                (
                    "exclude",
                    models.ManyToManyField(
                        blank=True,
                        help_text="Subscriptions to exclude from mailing, all other subscribers will receive message",
                        to="mailinglist.subscription",
                    ),
                ),
                (
                    "message",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="submission",
                        to="mailinglist.message",
                    ),
                ),
                (
                    "sendings",
                    models.ManyToManyField(
                        related_name="sendings",
                        through="mailinglist.Sending",
                        to="mailinglist.subscription",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="sending",
            name="submission",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT, to="mailinglist.submission"
            ),
        ),
        migrations.AddField(
            model_name="sending",
            name="subscription",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                to="mailinglist.subscription",
            ),
        ),
        migrations.CreateModel(
            name="MessagePart",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("heading", models.CharField(max_length=128)),
                ("order", models.PositiveSmallIntegerField()),
                ("text", models.TextField()),
                (
                    "message",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="message_parts",
                        to="mailinglist.message",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="GlobalDeny",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True)),
                (
                    "user",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="mailinglist_deny",
                        to=settings.MAILINGLIST_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.AddConstraint(
            model_name="subscription",
            constraint=models.UniqueConstraint(
                fields=("user", "mailing_list"), name="unique_user_per_mailing_list"
            ),
        ),
        migrations.AddConstraint(
            model_name="messagepart",
            constraint=models.UniqueConstraint(
                fields=("message", "order"), name="unique_order_per_message"
            ),
        ),
        migrations.AddConstraint(
            model_name="message",
            constraint=models.UniqueConstraint(
                fields=("slug", "mailing_list"), name="unique_slug_per_mailing_list"
            ),
        ),
    ]



from unittest.mock import patch
import pytest
from django.forms import ValidationError
from django.core.files.base import ContentFile
from mailinglist.admin_forms import ImportForm, ConfirmForm, SubmissionModelForm


class TestImportForm:
    def test_clean_missing_data_address_file(self, mailing_list):
        _data = {"ignore_errors": True, "mailing_list": mailing_list}
        form = ImportForm()
        form.cleaned_data = _data
        ret = form.clean()
        assert ret == _data
        assert not hasattr(form, "addresses")

    def test_clean_missing_data_ignore_errors(self, mailing_list):
        _data = {"address_file": True, "mailing_list": mailing_list}
        form = ImportForm()
        form.cleaned_data = _data
        ret = form.clean()
        assert not hasattr(form, "addresses")

    def test_clean_missing_data_ignore_errors(self):
        _data = {"address_file": True, "ignore_errors": True}
        form = ImportForm()
        form.cleaned_data = _data
        ret = form.clean()
        assert not hasattr(form, "addresses")

    def test_clean_bad_file_extension(self, mailing_list):
        _file = ContentFile(
            "first_name,last_name,email\nSeymore,Lastman,last@email.com", "somefile.bat"
        )
        _file.content_type = "text/csv"
        _data = {
            "address_file": _file,
            "ignore_errors": True,
            "mailing_list": mailing_list,
        }
        form = ImportForm()
        form.cleaned_data = _data
        with pytest.raises(ValidationError):
            ret = form.clean()
        assert not hasattr(form, "addresses")

    def test_clean_bad_file_content_type(self, mailing_list):
        _file = ContentFile(
            "first_name,last_name,email\nSeymore,Lastman,last@email.com", "somefile.csv"
        )
        _file.content_type = "text/excel"
        _data = {
            "address_file": _file,
            "ignore_errors": True,
            "mailing_list": mailing_list,
        }
        form = ImportForm()
        form.cleaned_data = _data
        with pytest.raises(ValidationError):
            ret = form.clean()
        assert not hasattr(form, "addresses")

    @patch("mailinglist.admin_forms.parse_csv")
    def test_clean_no_addresses(self, p_parse_csv, address_file, mailing_list):
        _data = {
            "address_file": address_file,
            "ignore_errors": True,
            "mailing_list": mailing_list,
        }
        form = ImportForm()

        form.cleaned_data = _data
        p_parse_csv.return_value = {}
        with pytest.raises(ValidationError):
            ret = form.clean()
        assert form.addresses == {}

    @patch("mailinglist.admin_forms.parse_csv")
    def test_clean(self, p_parse_csv, address_file, mailing_list):
        _data = {
            "address_file": address_file,
            "ignore_errors": True,
            "mailing_list": mailing_list,
        }
        form = ImportForm()

        form.cleaned_data = _data
        p_parse_csv.return_value = {"data": "values"}
        ret = form.clean()
        assert form.addresses == {"data": "values"}

    def test_get_addresses_default(self):
        form = ImportForm()
        assert form.get_addresses() == {}

    def test_get_addresses(self):
        form = ImportForm()
        form.addresses = {"data": "values"}
        assert form.get_addresses() == {"data": "values"}


class TestConfirmForm:
    def test_clean(self):
        form = ConfirmForm()
        form.cleaned_data = {"confirm": True}
        ret = form.clean()
        assert ret == {"confirm": True}

    def test_clean_bad(self):
        form = ConfirmForm()
        form.cleaned_data = {"confirm": False}
        with pytest.raises(ValidationError):
            ret = form.clean()


class TestSubmissionModelForm:
    def test_init_new(self, message):
        form = SubmissionModelForm()
        assert list(form.fields["message"].queryset) == [message]

    def test_init_no_messages(self, message, submission):
        form = SubmissionModelForm()
        assert list(form.fields["message"].queryset) == []

    def test_init_with_submission(self, message, submission, subscription):
        form = SubmissionModelForm(instance=submission)
        assert list(form.fields["message"].queryset) == [message]
        assert list(form.fields["exclude"].queryset) == [subscription]



import pytest
from mailinglist.conf import MailinglistAppConf
from django.core.exceptions import ImproperlyConfigured


class TestMailinglistAppConf:
    def test_configure_default_sender_email(self):
        assert (
            MailinglistAppConf().configure_default_sender_email("something")
            == "something"
        )

    def test_configure_default_sender_email_fails(self):
        with pytest.raises(ImproperlyConfigured):
            MailinglistAppConf().configure_default_sender_email(None)

    def test_configure_base_url(self):
        assert MailinglistAppConf().configure_base_url("something") == "something"

    def test_configure_base_url_fails(self):
        with pytest.raises(ImproperlyConfigured):
            MailinglistAppConf().configure_base_url(None)



from unittest.mock import patch
from mailinglist.models import hookset_validation_wrapper


def test_subscription_string(subscription):
    assert str(subscription) == f"{subscription.user} on {subscription.mailing_list}"


def test_message_string(message):
    assert str(message) == message.title


def test_message_part_html(message_part):
    assert "<h1>This should render gloriously!</h1>" in message_part.html_text
    assert "<em>verbosely</em>" in message_part.html_text
    assert "<strong>exquisite</strong>" in message_part.html_text


def test_submission_string(submission):
    assert (
        str(submission) == f"{submission.message} to {submission.message.mailing_list}"
    )


def test_message_attachment_string(message_attachment):
    assert (
        str(message_attachment)
        == f"{message_attachment.filename} on {message_attachment.message.title}"
    )


def test_message_attachment_file_name(message_attachment):
    assert message_attachment.filename == "testing.jpg"


def test_message_attachment_save(message_attachment):
    message_attachment.filename = "something_else.jpg"
    message_attachment.save()
    message_attachment.refresh_from_db()
    assert message_attachment.filename == "something_else.jpg"


@patch("mailinglist.models.hookset.message_attachment_file_validator")
def test_hookset_validation_wrapper(p_validator):
    hookset_validation_wrapper(True)
    p_validator.assert_called_once_with(True)
